import time
import warnings
from copy import copy
from dataclasses import asdict, dataclass, replace
from typing import Any, Callable, cast

import numpy as np
from numpy.typing import NDArray
from typing_extensions import Self

from optimagic.differentiation.derivatives import first_derivative
from optimagic.differentiation.numdiff_options import NumdiffOptions
from optimagic.exceptions import UserFunctionRuntimeError, get_traceback
from optimagic.logging.logger import LogStore
from optimagic.logging.types import IterationState
from optimagic.optimization.fun_value import SpecificFunctionValue
from optimagic.optimization.history import History, HistoryEntry
from optimagic.parameters.bounds import Bounds
from optimagic.parameters.conversion import Converter
from optimagic.typing import (
    AggregationLevel,
    BatchEvaluator,
    Direction,
    ErrorHandling,
    EvalTask,
    PyTree,
)


@dataclass(frozen=True)
class InternalBounds(Bounds):
    lower: NDArray[np.float64] | None
    upper: NDArray[np.float64] | None
    soft_lower: None = None
    soft_upper: None = None


class InternalOptimizationProblem:
    def __init__(
        self,
        fun: Callable[[PyTree], SpecificFunctionValue],
        jac: Callable[[PyTree], PyTree] | None,
        fun_and_jac: Callable[[PyTree], tuple[SpecificFunctionValue, PyTree]] | None,
        converter: Converter,
        solver_type: AggregationLevel,
        direction: Direction,
        bounds: InternalBounds,
        numdiff_options: NumdiffOptions,
        error_handling: ErrorHandling,
        error_penalty_func: Callable[
            [NDArray[np.float64]],
            tuple[SpecificFunctionValue, NDArray[np.float64]],
        ],
        batch_evaluator: BatchEvaluator,
        linear_constraints: list[dict[str, Any]] | None,
        nonlinear_constraints: list[dict[str, Any]] | None,
        logger: LogStore[Any, Any] | None,
        # TODO: add hess and hessp
    ):
        self._fun = fun
        self._jac = jac
        self._fun_and_jac = fun_and_jac
        self._converter = converter
        self._solver_type = solver_type
        self._direction = direction
        self._bounds = bounds
        self._numdiff_options = numdiff_options
        self._error_handling = error_handling
        self._error_penalty_func = error_penalty_func
        self._batch_evaluator = batch_evaluator
        self._history = History()
        self._linear_constraints = linear_constraints
        self._nonlinear_constraints = nonlinear_constraints
        self._logger = logger
        self._step_id: int | None = None

    # ==================================================================================
    # Public methods used by optimizers
    # ==================================================================================

    def fun(self, x: NDArray[np.float64]) -> float | NDArray[np.float64]:
        fun_value, hist_entry = self._evaluate_fun(x)
        self._history.add_entry(hist_entry)
        return fun_value

    def jac(self, x: NDArray[np.float64]) -> NDArray[np.float64]:
        jac_value, hist_entry = self._evaluate_jac(x)
        self._history.add_entry(hist_entry)
        return jac_value

    def fun_and_jac(
        self, x: NDArray[np.float64]
    ) -> tuple[float | NDArray[np.float64], NDArray[np.float64]]:
        fun_and_jac_value, hist_entry = self._evaluate_fun_and_jac(x)
        self._history.add_entry(hist_entry)
        return fun_and_jac_value

    def batch_fun(
        self,
        x_list: list[NDArray[np.float64]],
        n_cores: int,
        batch_size: int | None = None,
    ) -> list[float | NDArray[np.float64]]:
        batch_size = n_cores if batch_size is None else batch_size
        batch_result = self._batch_evaluator(
            func=self._evaluate_fun,
            arguments=x_list,
            n_cores=n_cores,
            # This should always be raise because errors are already handled
            error_handling="raise",
        )
        fun_values = [result[0] for result in batch_result]
        hist_entries = [result[1] for result in batch_result]
        self._history.add_batch(hist_entries, batch_size)

        return fun_values

    def batch_jac(
        self,
        x_list: list[NDArray[np.float64]],
        n_cores: int,
        batch_size: int | None = None,
    ) -> list[NDArray[np.float64]]:
        batch_size = n_cores if batch_size is None else batch_size

        batch_result = self._batch_evaluator(
            func=self._evaluate_jac,
            arguments=x_list,
            n_cores=n_cores,
            # This should always be raise because errors are already handled
            error_handling="raise",
        )
        jac_values = [result[0] for result in batch_result]
        hist_entries = [result[1] for result in batch_result]
        self._history.add_batch(hist_entries, batch_size)
        return jac_values

    def batch_fun_and_jac(
        self,
        x_list: list[NDArray[np.float64]],
        n_cores: int,
        batch_size: int | None = None,
    ) -> list[tuple[float | NDArray[np.float64], NDArray[np.float64]]]:
        batch_size = n_cores if batch_size is None else batch_size
        batch_result = self._batch_evaluator(
            func=self._evaluate_fun_and_jac,
            arguments=x_list,
            n_cores=n_cores,
            # This should always be raise because errors are already handled
            error_handling="raise",
        )
        fun_and_jac_values = [result[0] for result in batch_result]
        hist_entries = [result[1] for result in batch_result]
        self._history.add_batch(hist_entries, batch_size)

        return fun_and_jac_values

    def exploration_fun(
        self,
        x_list: list[NDArray[np.float64]],
        n_cores: int,
        batch_size: int | None = None,
    ) -> list[float]:
        batch_size = n_cores if batch_size is None else batch_size
        batch_result = self._batch_evaluator(
            func=self._evaluate_exploration_fun,
            arguments=x_list,
            n_cores=n_cores,
            # This should always be raise because errors are already handled
            error_handling="raise",
        )
        fun_values = [result[0] for result in batch_result]
        hist_entries = [result[1] for result in batch_result]
        self._history.add_batch(hist_entries, batch_size)

        return fun_values

    def with_new_history(self) -> Self:
        new = copy(self)
        new._history = History()
        return new

    def with_error_handling(self, error_handling: ErrorHandling) -> Self:
        new = copy(self)
        new._error_handling = error_handling
        return new

    def with_step_id(self, step_id: int) -> Self:
        new = copy(self)
        new._step_id = step_id
        return new

    # ==================================================================================
    # Public attributes
    # ==================================================================================

    @property
    def linear_constraints(self) -> list[dict[str, Any]] | None:
        return self._linear_constraints

    @property
    def nonlinear_constraints(self) -> list[dict[str, Any]] | None:
        return self._nonlinear_constraints

    @property
    def direction(self) -> Direction:
        return self._direction

    @property
    def history(self) -> History:
        return self._history

    @property
    def bounds(self) -> InternalBounds:
        return self._bounds

    @property
    def logger(self) -> LogStore[Any, Any] | None:
        return self._logger

    # ==================================================================================
    # Implementation of the public functions; The main difference is that the lower-
    # level implementations return a history entry instead of adding it to the history
    # directly so they can be called in parallel!
    # ==================================================================================

    def _evaluate_fun(
        self, x: NDArray[np.float64]
    ) -> tuple[float | NDArray[np.float64], HistoryEntry]:
        fun_value, hist_entry, log_entry = self._pure_evaluate_fun(x)

        if self._logger:
            self._logger.iteration_store.insert(log_entry)

        return fun_value, hist_entry

    def _evaluate_jac(
        self, x: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], HistoryEntry]:
        if self._jac is not None:
            jac_value, hist_entry, log_entry = self._pure_evaluate_jac(x)
        else:
            if self._fun_and_jac is not None:
                (_, jac_value), hist_entry, log_entry = self._pure_evaluate_fun_and_jac(
                    x
                )
            else:
                (_, jac_value), hist_entry, log_entry = (
                    self._pure_evaluate_numerical_fun_and_jac(x)
                )

            hist_entry = replace(hist_entry, task=EvalTask.JAC)

        if self._logger:
            self._logger.iteration_store.insert(log_entry)

        return jac_value, hist_entry

    def _evaluate_exploration_fun(
        self, x: NDArray[np.float64]
    ) -> tuple[float, HistoryEntry]:
        fun_value, hist_entry, log_entry = self._pure_exploration_fun(x)

        if self._logger:
            self._logger.iteration_store.insert(log_entry)

        return fun_value, hist_entry

    def _evaluate_fun_and_jac(
        self, x: NDArray[np.float64]
    ) -> tuple[tuple[float | NDArray[np.float64], NDArray[np.float64]], HistoryEntry]:
        if self._fun_and_jac is not None:
            (fun_value, jac_value), hist_entry, log_entry = (
                self._pure_evaluate_fun_and_jac(x)
            )
        elif self._jac is not None:
            fun_value, hist_entry, log_entry_fun = self._pure_evaluate_fun(x)
            jac_value, _, log_entry_jac = self._pure_evaluate_jac(x)
            hist_entry = replace(hist_entry, task=EvalTask.FUN_AND_JAC)
            log_entry = log_entry_fun.combine(log_entry_jac)
        else:
            (fun_value, jac_value), hist_entry, log_entry = (
                self._pure_evaluate_numerical_fun_and_jac(x)
            )

        if self._logger:
            self._logger.iteration_store.insert(log_entry)

        return (fun_value, jac_value), hist_entry

    # ==================================================================================
    # Atomic evaluations of user provided functions or numerical derivatives
    # ==================================================================================

    def _pure_evaluate_fun(
        self, x: NDArray[np.float64]
    ) -> tuple[float | NDArray[np.float64], HistoryEntry, IterationState]:
        """Evaluate fun and handle exceptions.

        This function does all the conversions from x to params and from
        SpecificFunctionValue to the internal value, including a sign flip for
        maximization.

        If any exception occurs during the evaluation of fun and error handling is set
        to CONTINUE, the fun value is replaced by a penalty value and a warning is
        issued.

        """
        now = time.perf_counter()
        params = self._converter.params_from_internal(x)
        traceback: None | str = None
        try:
            fun_value = self._fun(params)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            if self._error_handling in (
                ErrorHandling.RAISE,
                ErrorHandling.RAISE_STRICT,
            ):
                msg = "An error occurred when evaluating fun during optimization."
                raise UserFunctionRuntimeError(msg) from e
            else:
                traceback = get_traceback()
                msg = (
                    "The following exception was caught when evaluating fun during "
                    "optimization. The fun value was replaced by a penalty value to "
                    f"continue with the optimization.:\n\n{traceback}"
                )
                warnings.warn(msg)
                fun_value, _ = self._error_penalty_func(x)

        algo_fun_value, hist_fun_value = _process_fun_value(
            value=fun_value, solver_type=self._solver_type, direction=self._direction
        )

        hist_entry = HistoryEntry(
            params=params,
            fun=hist_fun_value,
            time=now,
            task=EvalTask.FUN,
        )

        log_entry = IterationState(
            params=params,
            timestamp=now,
            scalar_fun=hist_fun_value,
            valid=not bool(traceback),
            raw_fun=fun_value,
            step=self._step_id,
            exceptions=traceback,
        )

        return algo_fun_value, hist_entry, log_entry

    def _pure_evaluate_jac(
        self, x: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], HistoryEntry, IterationState]:
        if self._jac is None:
            raise ValueError("The jac function is not defined.")

        now = time.perf_counter()
        traceback: None | str = None

        params = self._converter.params_from_internal(x)
        try:
            jac_value = self._jac(params)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            if self._error_handling in (
                ErrorHandling.RAISE,
                ErrorHandling.RAISE_STRICT,
            ):
                msg = "An error occurred when evaluating jac during optimization."
                raise UserFunctionRuntimeError(msg) from e
            else:
                traceback = get_traceback()

                msg = (
                    "The following exception was caught when evaluating jac during "
                    "optimization. The jac value was replaced by a penalty value to "
                    f"continue with the optimization.:\n\n{traceback}"
                )
                warnings.warn(msg)
                _, jac_value = self._error_penalty_func(x)

        out_jac = _process_jac_value(
            value=jac_value, direction=self._direction, converter=self._converter, x=x
        )

        hist_entry = HistoryEntry(
            params=params,
            fun=None,
            time=now,
            task=EvalTask.JAC,
        )

        log_entry = IterationState(
            params=params,
            timestamp=now,
            scalar_fun=None,
            valid=not bool(traceback),
            raw_fun=None,
            step=self._step_id,
            exceptions=traceback,
        )

        return out_jac, hist_entry, log_entry

    def _pure_evaluate_numerical_fun_and_jac(
        self, x: NDArray[np.float64]
    ) -> tuple[
        tuple[float | NDArray[np.float64], NDArray[np.float64]],
        HistoryEntry,
        IterationState,
    ]:
        now = time.perf_counter()
        traceback: None | str = None

        def func(x: NDArray[np.float64]) -> SpecificFunctionValue:
            p = self._converter.params_from_internal(x)
            return self._fun(p)

        try:
            numdiff_res = first_derivative(
                func,
                x,
                bounds=self._bounds,
                **asdict(self._numdiff_options),
                unpacker=lambda x: x.internal_value(self._solver_type),
                error_handling="raise_strict",
            )
            fun_value = numdiff_res.func_value
            jac_value = numdiff_res.derivative
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            if self._error_handling in (
                ErrorHandling.RAISE,
                ErrorHandling.RAISE_STRICT,
            ):
                msg = (
                    "An error occurred when evaluating a numerical derivative "
                    "during optimization."
                )
                raise UserFunctionRuntimeError(msg) from e
            else:
                traceback = get_traceback()

                msg = (
                    "The following exception was caught when calculating a "
                    "numerical derivative during optimization. The jac value was "
                    "replaced by a penalty value to continue with the optimization."
                    f":\n\n{traceback}"
                )
                warnings.warn(msg)
                fun_value, jac_value = self._error_penalty_func(x)

        algo_fun_value, hist_fun_value = _process_fun_value(
            value=fun_value,  # type: ignore
            solver_type=self._solver_type,
            direction=self._direction,
        )

        if self._direction == Direction.MAXIMIZE:
            jac_value = -jac_value

        hist_entry = HistoryEntry(
            params=self._converter.params_from_internal(x),
            fun=hist_fun_value,
            time=now,
            task=EvalTask.FUN_AND_JAC,
        )

        log_entry = IterationState(
            params=self._converter.params_from_internal(x),
            timestamp=now,
            scalar_fun=hist_fun_value,
            valid=not bool(traceback),
            raw_fun=fun_value,
            step=self._step_id,
            exceptions=traceback,
        )

        return (algo_fun_value, jac_value), hist_entry, log_entry

    def _pure_exploration_fun(
        self, x: NDArray[np.float64]
    ) -> tuple[float, HistoryEntry, IterationState]:
        now = time.perf_counter()
        params = self._converter.params_from_internal(x)
        traceback: None | str = None

        try:
            fun_value = self._fun(params)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            traceback = get_traceback()

            msg = (
                "The following exception was caught when evaluating fun during the "
                "exploration phase of a multistart optimization. The fun value was "
                "replaced by a penalty value to continue with the "
                f"optimization.:\n\n{traceback}"
            )
            warnings.warn(msg)
            fun_value, _ = self._error_penalty_func(x)

        if not traceback:
            algo_fun_value, hist_fun_value = _process_fun_value(
                value=fun_value,
                # For exploration we always need a scalar value
                solver_type=AggregationLevel.SCALAR,
                direction=self._direction,
            )
        else:
            algo_fun_value = -np.inf
            hist_fun_value = -np.inf
            if self._direction == Direction.MAXIMIZE:
                hist_fun_value = np.inf

        hist_entry = HistoryEntry(
            params=params,
            fun=hist_fun_value,
            time=now,
            task=EvalTask.EXPLORATION,
        )

        log_entry = IterationState(
            params=params,
            timestamp=now,
            scalar_fun=hist_fun_value,
            valid=not bool(traceback),
            raw_fun=fun_value,
            step=self._step_id,
            exceptions=traceback,
        )

        return cast(float, algo_fun_value), hist_entry, log_entry

    def _pure_evaluate_fun_and_jac(
        self, x: NDArray[np.float64]
    ) -> tuple[
        tuple[float | NDArray[np.float64], NDArray[np.float64]],
        HistoryEntry,
        IterationState,
    ]:
        if self._fun_and_jac is None:
            raise ValueError("The fun_and_jac function is not defined.")

        now = time.perf_counter()
        traceback: None | str = None
        params = self._converter.params_from_internal(x)

        try:
            fun_value, jac_value = self._fun_and_jac(params)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            if self._error_handling in (
                ErrorHandling.RAISE,
                ErrorHandling.RAISE_STRICT,
            ):
                msg = (
                    "An error occurred when evaluating fun_and_jac during optimization."
                )
                raise UserFunctionRuntimeError(msg) from e
            else:
                traceback = get_traceback()
                msg = (
                    "The following exception was caught when evaluating fun_and_jac "
                    "during optimization. The fun and jac values were replaced by "
                    f"penalty values to continue with the optimization.:\n\n{traceback}"
                )
                warnings.warn(msg)

                fun_value, jac_value = self._error_penalty_func(x)

        algo_fun_value, hist_fun_value = _process_fun_value(
            value=fun_value, solver_type=self._solver_type, direction=self._direction
        )

        if traceback:
            out_jac = jac_value
        else:
            out_jac = self._converter.derivative_to_internal(jac_value, x)

        if self._direction == Direction.MAXIMIZE:
            out_jac = -out_jac

        hist_entry = HistoryEntry(
            params=params,
            fun=hist_fun_value,
            time=now,
            task=EvalTask.FUN_AND_JAC,
        )

        log_entry = IterationState(
            params=params,
            timestamp=now,
            scalar_fun=hist_fun_value,
            valid=not bool(traceback),
            raw_fun=fun_value,
            step=self._step_id,
            exceptions=traceback,
        )

        return (algo_fun_value, out_jac), hist_entry, log_entry


def _process_fun_value(
    value: SpecificFunctionValue,
    solver_type: AggregationLevel,
    direction: Direction,
) -> tuple[float | NDArray[np.float64], float]:
    """Post-process a function value for use by the algorithm and as history entry.

    The sign flip for maximization is only applied to the value that will be passed to
    the algorithm.

    Args:
        value: The function value.
        solver_type: The aggregation level of the solver.
        direction: The direction of optimization.

    Returns:
        A tuple of the function value for the algorithm and the function value for the
        history entry.

    """
    algo_value = value.internal_value(solver_type)
    history_value = cast(float, value.internal_value(AggregationLevel.SCALAR))
    if direction == Direction.MAXIMIZE:
        algo_value = -algo_value

    return algo_value, history_value


def _process_jac_value(
    value: SpecificFunctionValue,
    direction: Direction,
    converter: Converter,
    x: NDArray[np.float64],
) -> NDArray[np.float64]:
    """Post-process a for use by the algorithm.

    Args:
        value: The Jacobian value.
        direction: The direction of optimization.
        converter: The converter object.

    Returns:
        The Jacobian value for the algorithm.

    """

    out_value = converter.derivative_to_internal(value, x)
    if direction == Direction.MAXIMIZE:
        out_value = -out_value

    return out_value



from optimagic.batch_evaluators import joblib_batch_evaluator as _joblib_batch_evaluator
from optimagic.batch_evaluators import (
    pathos_mp_batch_evaluator as _pathos_mp_batch_evaluator,
)
from optimagic.batch_evaluators import (
    process_batch_evaluator as _process_batch_evaluator,
)
from optimagic.decorators import deprecated

MSG = (
    "estimagic.batch_evaluators.{name} has been deprecated in version 0.5.0. Use "
    "optimagic.batch_evaluators.{name} instead. This function will be removed in "
    "version 0.6.0."
)


pathos_mp_batch_evaluator = deprecated(
    _pathos_mp_batch_evaluator, MSG.format(name="pathos_mp_batch_evaluator")
)

joblib_batch_evaluator = deprecated(
    _joblib_batch_evaluator, MSG.format(name="joblib_batch_evaluator")
)

process_batch_evaluator = deprecated(
    _process_batch_evaluator, MSG.format(name="process_batch_evaluator")
)



import warnings
from dataclasses import dataclass
from typing import Any, Dict, Optional

import numpy as np
import pandas as pd

from optimagic import deprecations
from optimagic.logging.logger import LogReader
from optimagic.optimization.history import History
from optimagic.shared.compat import pd_df_map
from optimagic.typing import PyTree
from optimagic.utilities import to_pickle


@dataclass
class OptimizeResult:
    """Optimization result object.

    **Attributes**

    Attributes:
        params: The optimal parameters.
        fun: The optimal criterion value.
        start_fun: The criterion value at the start parameters.
        start_params: The start parameters.
        algorithm: The algorithm used for the optimization.
        direction: Maximize or minimize.
        n_free: Number of free parameters.
        message: Message returned by the underlying algorithm.
        success: Whether the optimization was successful.
        n_fun_evals: Number of criterion evaluations.
        n_jac_evals: Number of derivative evaluations.
        n_iterations: Number of iterations until termination.
        history: Optimization history.
        convergence_report: The convergence report.
        multistart_info: Multistart information.
        algorithm_output: Additional algorithm specific information.

    """

    params: Any
    fun: float
    start_fun: float
    start_params: Any
    algorithm: str
    direction: str
    n_free: int

    message: str | None = None
    success: bool | None = None
    n_fun_evals: int | None = None
    n_jac_evals: int | None = None
    n_hess_evals: int | None = None
    n_iterations: int | None = None
    status: int | None = None
    jac: PyTree | None = None
    hess: PyTree | None = None
    hess_inv: PyTree | None = None
    max_constraint_violation: float | None = None

    history: History | None = None

    convergence_report: Dict | None = None

    multistart_info: Optional["MultistartInfo"] = None
    algorithm_output: Dict[str, Any] | None = None
    logger: LogReader | None = None

    # ==================================================================================
    # Deprecations
    # ==================================================================================

    @property
    def criterion(self) -> float:
        msg = "The criterion attribute is deprecated. Use the fun attribute instead."
        warnings.warn(msg, FutureWarning)
        return self.fun

    @property
    def start_criterion(self) -> float:
        msg = (
            "The start_criterion attribute is deprecated. Use the start_fun attribute "
            "instead."
        )
        warnings.warn(msg, FutureWarning)
        return self.start_fun

    @property
    def n_criterion_evaluations(self) -> int | None:
        msg = (
            "The n_criterion_evaluations attribute is deprecated. Use the n_fun_evals "
            "attribute instead."
        )
        warnings.warn(msg, FutureWarning)
        return self.n_fun_evals

    @property
    def n_derivative_evaluations(self) -> int | None:
        msg = (
            "The n_derivative_evaluations attribute is deprecated. Use the n_jac_evals "
            "attribute instead."
        )
        warnings.warn(msg, FutureWarning)
        return self.n_jac_evals

    # ==================================================================================
    # Scipy aliases
    # ==================================================================================

    @property
    def x(self) -> PyTree:
        return self.params

    @property
    def x0(self) -> PyTree:
        return self.start_params

    @property
    def nfev(self) -> int | None:
        return self.n_fun_evals

    @property
    def nit(self) -> int | None:
        return self.n_iterations

    @property
    def njev(self) -> int | None:
        return self.n_jac_evals

    @property
    def nhev(self) -> int | None:
        return self.n_hess_evals

    # Enable attribute access using dictionary-style notation for scipy compatibility
    def __getitem__(self, key):
        return getattr(self, key)

    def __repr__(self) -> str:
        first_line = (
            f"{self.direction.title()} with {self.n_free} free parameters terminated"
        )

        if self.success is not None:
            snippet = "successfully" if self.success else "unsuccessfully"
            first_line += f" {snippet}"

        counters = [
            ("criterion evaluations", self.n_fun_evals),
            ("derivative evaluations", self.n_jac_evals),
            ("iterations", self.n_iterations),
        ]

        counters = [(n, v) for n, v in counters if v is not None]

        if counters:
            name, val = counters[0]
            counter_msg = f"after {val} {name}"
            if len(counters) >= 2:
                for name, val in counters[1:-1]:
                    counter_msg += f", {val} {name}"

                name, val = counters[-1]
                counter_msg += f" and {val} {name}"
            first_line += f" {counter_msg}"

        first_line += "."

        if self.message:
            message = f"The {self.algorithm} algorithm reported: {self.message}"
        else:
            message = None

        if self.start_fun is not None and self.fun is not None:
            improvement = (
                f"The value of criterion improved from {self.start_fun} to "
                f"{self.fun}."
            )
        else:
            improvement = None

        if self.convergence_report is not None:
            convergence = _format_convergence_report(
                self.convergence_report, self.algorithm
            )
        else:
            convergence = None

        sections = [first_line, improvement, message, convergence]
        sections = [sec for sec in sections if sec is not None]

        msg = "\n\n".join(sections)

        return msg

    def to_pickle(self, path):
        """Save the OptimizeResult object to pickle.

        Args:
            path (str, pathlib.Path): A str or pathlib.path ending in .pkl or .pickle.

        """
        to_pickle(self, path=path)


@dataclass(frozen=True)
class MultistartInfo:
    """Information about the multistart optimization.

    Attributes:
        start_parameters: List of start parameters for each optimization.
        local_optima: List of optimization results.
        exploration_sample: List of parameters used for exploration.
        exploration_results: List of function values corresponding to exploration.
        n_optimizations: Number of local optimizations that were run.

    """

    start_parameters: list[PyTree]
    local_optima: list[OptimizeResult]
    exploration_sample: list[PyTree]
    exploration_results: list[float]

    def __getitem__(self, key):
        deprecations.throw_dict_access_future_warning(key, obj_name=type(self).__name__)
        return getattr(self, key)

    @property
    def n_optimizations(self) -> int:
        return len(self.local_optima)


def _format_convergence_report(report, algorithm):
    report = pd.DataFrame.from_dict(report)
    columns = ["one_step", "five_steps"]

    table = pd_df_map(report[columns], _format_float).astype(str)

    for col in "one_step", "five_steps":
        table[col] = table[col] + _create_stars(report[col])

    table = table.to_string(justify="center")

    introduction = (
        f"Independent of the convergence criteria used by {algorithm}, "
        "the strength of convergence can be assessed by the following criteria:"
    )

    explanation = (
        "(***: change <= 1e-10, **: change <= 1e-8, *: change <= 1e-5. "
        "Change refers to a change between accepted steps. The first column only "
        "considers the last step. The second column considers the last five steps.)"
    )

    out = "\n\n".join([introduction, table, explanation])

    return out


def _create_stars(sr):
    stars = pd.cut(
        sr,
        bins=[-np.inf, 1e-10, 1e-8, 1e-5, np.inf],
        labels=["***", "** ", "*  ", "   "],
    ).astype("str")

    return stars


def _format_float(number):
    """Round to four significant digits."""
    return f"{number:.4g}"



from typing import Callable

import numpy as np
from numpy.typing import NDArray

from optimagic.config import CRITERION_PENALTY_CONSTANT, CRITERION_PENALTY_SLOPE
from optimagic.optimization.fun_value import (
    LeastSquaresFunctionValue,
    LikelihoodFunctionValue,
    ScalarFunctionValue,
    SpecificFunctionValue,
)
from optimagic.typing import AggregationLevel, Direction


def _scalar_penalty(
    x: NDArray[np.float64],
    constant: float | NDArray[np.float64],
    slope: float | NDArray[np.float64],
    x0: NDArray[np.float64],
    dim_out: int | None = None,
) -> tuple[ScalarFunctionValue, NDArray[np.float64]]:  # noqa: ARG001
    value = constant + slope * np.linalg.norm(x - x0)
    jac = slope * (x - x0) / np.linalg.norm(x - x0)
    return ScalarFunctionValue(value=value), jac


def _likelihood_penalty(
    x: NDArray[np.float64],
    constant: float | NDArray[np.float64],
    slope: float | NDArray[np.float64],
    x0: NDArray[np.float64],
    dim_out: int,
) -> tuple[LikelihoodFunctionValue, NDArray[np.float64]]:  # noqa: ARG001
    factor = (constant + slope * np.linalg.norm(x - x0)) / dim_out
    contrib = np.ones(dim_out) * factor
    row = slope * (x - x0) / (dim_out * np.linalg.norm(x - x0))
    jac = np.full((dim_out, len(x)), row)
    return LikelihoodFunctionValue(value=contrib), jac


def _penalty_residuals(
    x: NDArray[np.float64],
    constant: float | NDArray[np.float64],
    slope: float | NDArray[np.float64],
    x0: NDArray[np.float64],
    dim_out: int,
) -> tuple[LeastSquaresFunctionValue, NDArray[np.float64]]:
    factor = np.sqrt((constant + slope * np.linalg.norm(x - x0)) / dim_out)
    contrib = np.ones(dim_out) * factor

    scalar_penalty, _ = _scalar_penalty(x, constant, slope, x0)
    inner_deriv = slope * (x - x0) / np.linalg.norm(x - x0)
    outer_deriv = 0.5 / np.sqrt(scalar_penalty.value * dim_out)
    row = outer_deriv * inner_deriv
    jac = np.full((dim_out, len(x)), row)

    return LeastSquaresFunctionValue(value=contrib), jac


def get_error_penalty_function(
    start_x: NDArray[np.float64],
    start_criterion: SpecificFunctionValue,
    error_penalty: dict[str, float] | None,
    solver_type: AggregationLevel,
    direction: Direction,
) -> Callable[[NDArray[np.float64]], tuple[SpecificFunctionValue, NDArray[np.float64]]]:
    error_penalty = {} if error_penalty is None else error_penalty

    first_value = start_criterion.internal_value(solver_type)

    constant, slope = _process_error_penalty(
        error_penalty=error_penalty,
        first_value=first_value,
        direction=direction,
    )

    dim_out = (
        1
        if solver_type == AggregationLevel.SCALAR
        else len(start_criterion.internal_value(solver_type))  # type: ignore
    )

    _penalty: Callable[
        [
            NDArray[np.float64],
            float | NDArray[np.float64],
            float | NDArray[np.float64],
            NDArray[np.float64],
            int,
        ],
        tuple[SpecificFunctionValue, NDArray[np.float64]],
    ]
    if solver_type == AggregationLevel.SCALAR:
        _penalty = _scalar_penalty
    elif solver_type == AggregationLevel.LIKELIHOOD:
        _penalty = _likelihood_penalty
    elif solver_type == AggregationLevel.LEAST_SQUARES:
        _penalty = _penalty_residuals

    def penalty(
        x: NDArray[np.float64],
    ) -> tuple[SpecificFunctionValue, NDArray[np.float64]]:
        out = _penalty(
            x=x,
            constant=constant,
            slope=slope,
            x0=start_x,
            dim_out=dim_out,
        )
        return out

    return penalty


def _process_error_penalty(
    error_penalty: dict[str, float] | None,
    first_value: float | NDArray[np.float64],
    direction: Direction,
) -> tuple[float | NDArray[np.float64], float | NDArray[np.float64]]:
    """Add default options to error_penalty options."""
    if error_penalty is not None:
        error_penalty = error_penalty.copy()
    else:
        error_penalty = {}

    if direction == Direction.MINIMIZE:
        default_constant = (
            first_value + np.abs(first_value) + CRITERION_PENALTY_CONSTANT
        )
        default_slope = CRITERION_PENALTY_SLOPE
    elif direction == Direction.MAXIMIZE:
        default_constant = (
            first_value - np.abs(first_value) - CRITERION_PENALTY_CONSTANT
        )
        default_slope = -CRITERION_PENALTY_SLOPE
    else:
        raise ValueError()

    constant = error_penalty.get("constant", default_constant)
    slope = error_penalty.get("slope", default_slope)

    return constant, slope



import functools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, ParamSpec

import numpy as np
from numpy.typing import NDArray
from pybaum import tree_just_flatten

from optimagic.exceptions import InvalidFunctionError
from optimagic.parameters.tree_registry import get_registry
from optimagic.typing import AggregationLevel, PyTree, Scalar
from optimagic.utilities import isscalar


@dataclass(frozen=True)
class FunctionValue:
    value: float | PyTree
    info: dict[str, Any] | None = None


class SpecificFunctionValue(FunctionValue, ABC):
    @abstractmethod
    def internal_value(
        self, solver_type: AggregationLevel
    ) -> float | NDArray[np.float64]:
        pass


@dataclass(frozen=True)
class ScalarFunctionValue(SpecificFunctionValue):
    value: Scalar
    info: dict[str, Any] | None = None

    def __post_init__(self) -> None:
        if not isscalar(self.value):
            raise InvalidFunctionError(
                f"Scalar objective values need to be scalars, not: {type(self.value)}. "
                "If you meant to provide a scalar objective function, make sure it "
                "returns a scalar value. If you meant to provide a least_squares or "
                "likelihood function, use the mark.least_squares or mark.likelihood "
                "decorators."
            )

    def internal_value(self, solver_type: AggregationLevel) -> float:
        if solver_type == AggregationLevel.SCALAR:
            val = float(self.value)
        else:
            raise InvalidFunctionError(
                f"You are using a {solver_type.value} optimizer but provided a "
                "scalar objective function."
            )
        return val


@dataclass(frozen=True)
class LeastSquaresFunctionValue(SpecificFunctionValue):
    value: PyTree
    info: dict[str, Any] | None = None

    def __post_init__(self) -> None:
        if isscalar(self.value):
            raise InvalidFunctionError(
                "Least squares objective values cannot be scalars. Your value has "
                f"scalar type: {type(self.value)}. If you meant to provide a least "
                "squares objective function, make sure it does not have a scalar value."
                " If you meant to provide a scalar function, use the mark.scalar "
                "decorator."
            )

    def internal_value(
        self, solver_type: AggregationLevel
    ) -> float | NDArray[np.float64]:
        resid = _get_flat_value(self.value)

        val: float | NDArray[np.float64]

        if solver_type == AggregationLevel.LEAST_SQUARES:
            val = resid
        elif solver_type == AggregationLevel.LIKELIHOOD:
            val = resid**2
        else:
            val = float(resid @ resid)
        return val


@dataclass(frozen=True)
class LikelihoodFunctionValue(SpecificFunctionValue):
    value: PyTree
    info: dict[str, Any] | None = None

    def __post_init__(self) -> None:
        if isscalar(self.value):
            raise InvalidFunctionError(
                "Likelihood objective values cannot be scalars. Your value has scalar "
                f"type: {type(self.value)}. If you meant to provide a likelihood "
                "objective function, make sure it does not have a scalar value. If you "
                "meant to provide a scalar function, use the mark.scalar decorator."
            )

    def internal_value(
        self, solver_type: AggregationLevel
    ) -> float | NDArray[np.float64]:
        loglikes = _get_flat_value(self.value)

        val: float | NDArray[np.float64]

        if solver_type == AggregationLevel.LIKELIHOOD:
            val = loglikes
        elif solver_type == AggregationLevel.SCALAR:
            val = float(np.sum(loglikes))
        else:
            raise InvalidFunctionError(
                "You are using a least_squares optimizer but provided a "
                "likelihood objective function."
            )
        return val


def _get_flat_value(value: PyTree) -> NDArray[np.float64]:
    """Flatten a PyTree value to a 1d numpy array with multiple fast paths."""
    if isinstance(value, np.ndarray) and value.ndim == 1:
        flat = value
    elif isinstance(value, np.ndarray):
        flat = value.flatten()
    else:
        registry = get_registry(extended=True)
        flat = tree_just_flatten(value, registry=registry)

    flat_arr = np.asarray(flat, dtype=np.float64)
    return flat_arr


def convert_fun_output_to_function_value(
    raw: Scalar | PyTree | FunctionValue, problem_type: AggregationLevel
) -> SpecificFunctionValue:
    out: FunctionValue
    if problem_type == AggregationLevel.SCALAR:
        out = _convert_output_to_scalar_function_value(raw)
    elif problem_type == AggregationLevel.LEAST_SQUARES:
        out = _convert_output_to_least_squares_function_value(raw)
    elif problem_type == AggregationLevel.LIKELIHOOD:
        out = _convert_output_to_likelihood_function_value(raw)
    return out


def _convert_output_to_scalar_function_value(
    raw: Scalar | FunctionValue,
) -> ScalarFunctionValue:
    if isinstance(raw, ScalarFunctionValue):
        out = raw
    elif isinstance(raw, FunctionValue):
        out = ScalarFunctionValue(value=raw.value, info=raw.info)
    else:
        out = ScalarFunctionValue(value=raw)
    return out


def _convert_output_to_least_squares_function_value(
    raw: PyTree | FunctionValue,
) -> LeastSquaresFunctionValue:
    if isinstance(raw, LeastSquaresFunctionValue):
        out = raw
    elif isinstance(raw, FunctionValue):
        out = LeastSquaresFunctionValue(value=raw.value, info=raw.info)
    else:
        out = LeastSquaresFunctionValue(value=raw)
    return out


def _convert_output_to_likelihood_function_value(
    raw: PyTree | FunctionValue,
) -> LikelihoodFunctionValue:
    if isinstance(raw, LikelihoodFunctionValue):
        out = raw
    elif isinstance(raw, FunctionValue):
        out = LikelihoodFunctionValue(value=raw.value, info=raw.info)
    else:
        out = LikelihoodFunctionValue(value=raw)
    return out


P = ParamSpec("P")


def enforce_return_type(
    problem_type: AggregationLevel,
) -> Callable[
    [Callable[P, Scalar | PyTree | FunctionValue]], Callable[P, SpecificFunctionValue]
]:
    """Enforce a strict return type for objective functions based on problem_type.

    This has no effect if the function already returns the strictest possible type for
    the problem_type but converts everything else to that type.

    """

    def decorator_enforce(
        func: Callable[P, Scalar | PyTree | FunctionValue],
    ) -> Callable[P, SpecificFunctionValue]:
        if problem_type == AggregationLevel.SCALAR:

            @functools.wraps(func)
            def wrapper_enforce(
                *args: P.args, **kwargs: P.kwargs
            ) -> ScalarFunctionValue:
                raw = func(*args, **kwargs)
                return _convert_output_to_scalar_function_value(raw)
        elif problem_type == AggregationLevel.LEAST_SQUARES:

            @functools.wraps(func)
            def wrapper_enforce(
                *args: P.args, **kwargs: P.kwargs
            ) -> LeastSquaresFunctionValue:
                raw = func(*args, **kwargs)
                return _convert_output_to_least_squares_function_value(raw)
        elif problem_type == AggregationLevel.LIKELIHOOD:

            @functools.wraps(func)
            def wrapper_enforce(
                *args: P.args, **kwargs: P.kwargs
            ) -> LikelihoodFunctionValue:
                raw = func(*args, **kwargs)
                return _convert_output_to_likelihood_function_value(raw)

        return wrapper_enforce

    return decorator_enforce


def enforce_return_type_with_jac(
    problem_type: AggregationLevel,
) -> Callable[
    [Callable[P, tuple[Scalar | PyTree | FunctionValue, PyTree]]],
    Callable[P, tuple[SpecificFunctionValue, PyTree]],
]:
    """Enforce a strict return type for fun_and_jac based on problem_type.

    This has no effect if the first return value of the function already has the
    strictest possible type for the problem_type but converts everything else to that
    type. The second return value stays unchanged.

    """

    def decorator_enforce(
        func: Callable[P, tuple[Scalar | PyTree | FunctionValue, PyTree]],
    ) -> Callable[P, tuple[SpecificFunctionValue, PyTree]]:
        if problem_type == AggregationLevel.SCALAR:

            @functools.wraps(func)
            def wrapper_enforce(
                *args: P.args, **kwargs: P.kwargs
            ) -> tuple[ScalarFunctionValue, PyTree]:
                raw = func(*args, **kwargs)
                return (_convert_output_to_scalar_function_value(raw[0]), raw[1])
        elif problem_type == AggregationLevel.LEAST_SQUARES:

            @functools.wraps(func)
            def wrapper_enforce(
                *args: P.args, **kwargs: P.kwargs
            ) -> tuple[LeastSquaresFunctionValue, PyTree]:
                raw = func(*args, **kwargs)
                return (_convert_output_to_least_squares_function_value(raw[0]), raw[1])
        elif problem_type == AggregationLevel.LIKELIHOOD:

            @functools.wraps(func)
            def wrapper_enforce(
                *args: P.args, **kwargs: P.kwargs
            ) -> tuple[LikelihoodFunctionValue, PyTree]:
                raw = func(*args, **kwargs)
                return (_convert_output_to_likelihood_function_value(raw[0]), raw[1])

        return wrapper_enforce

    return decorator_enforce



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Sequence

import numpy as np
from numpy.typing import NDArray
from pybaum import leaf_names, tree_map
from pybaum import tree_just_flatten as tree_leaves
from scipy.optimize import Bounds as ScipyBounds

from optimagic.exceptions import InvalidBoundsError
from optimagic.parameters.tree_registry import get_registry
from optimagic.typing import PyTree, PyTreeRegistry


@dataclass(frozen=True)
class Bounds:
    lower: PyTree | None = None
    upper: PyTree | None = None
    soft_lower: PyTree | None = None
    soft_upper: PyTree | None = None


def pre_process_bounds(
    bounds: None | Bounds | ScipyBounds | Sequence[tuple[float, float]],
) -> Bounds | None:
    """Convert all valid types of specifying bounds to optimagic.Bounds.

    This just harmonizes multiple ways of specifying bounds into a single format.
    It does not check that bounds are valid or compatible with params.

    Args:
        bounds: The user provided bounds.

    Returns:
        The bounds in the optimagic format.

    Raises:
        InvalidBoundsError: If bounds cannot be processed, e.g. because they do not have
            the correct type.

    """
    if isinstance(bounds, ScipyBounds):
        bounds = Bounds(lower=bounds.lb, upper=bounds.ub)
    elif isinstance(bounds, Bounds) or bounds is None:
        pass
    else:
        try:
            bounds = _process_bounds_sequence(bounds)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            raise InvalidBoundsError(
                f"Invalid bounds of type: {type(bounds)}. Bounds must be "
                "optimagic.Bounds, scipy.optimize.Bounds or a Sequence of tuples with "
                "lower and upper bounds."
            ) from e
    return bounds


def _process_bounds_sequence(bounds: Sequence[tuple[float, float]]) -> Bounds:
    lower = np.full(len(bounds), -np.inf)
    upper = np.full(len(bounds), np.inf)

    for i, (lb, ub) in enumerate(bounds):
        if lb is not None:
            lower[i] = lb
        if ub is not None:
            upper[i] = ub
    return Bounds(lower=lower, upper=upper)


def get_internal_bounds(
    params: PyTree,
    bounds: Bounds | None = None,
    registry: PyTreeRegistry | None = None,
    add_soft_bounds: bool = False,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Create consolidated and flattened bounds for params.

    If params is a DataFrame with value column, the user provided bounds are
    extended with bounds from the params DataFrame.

    If no bounds are available the entry is set to minus np.inf for the lower bound and
    np.inf for the upper bound.

    The bounds provided in `bounds` override bounds provided in params if both are
    specified (in the case where params is a DataFrame with bounds as a column).

    Args:
        params: The parameter pytree.
        bounds: The lower and upper bounds.
        registry: pybaum registry.
        add_soft_bounds: If True, the element-wise maximum (minimum) of the lower and
            soft_lower (upper and soft_upper) bounds are taken. If False, the lower
            (upper) bounds are returned.

    Returns:
        Consolidated and flattened lower_bounds.
        Consolidated and flattened upper_bounds.

    """
    bounds = Bounds() if bounds is None else bounds

    fast_path = _is_fast_path(
        params=params,
        bounds=bounds,
        add_soft_bounds=add_soft_bounds,
    )
    if fast_path:
        return _get_fast_path_bounds(
            params=params,
            bounds=bounds,
        )

    registry = get_registry(extended=True) if registry is None else registry
    n_params = len(tree_leaves(params, registry=registry))

    # Fill leaves with np.nan. If params contains a data frame with bounds as a column,
    # that column is NOT overwritten (as long as an extended registry is used).
    nan_tree = tree_map(lambda leaf: np.nan, params, registry=registry)  # noqa: ARG005

    lower_flat = _update_bounds_and_flatten(nan_tree, bounds.lower, kind="lower_bound")
    upper_flat = _update_bounds_and_flatten(nan_tree, bounds.upper, kind="upper_bound")

    if len(lower_flat) != n_params:
        raise InvalidBoundsError("lower_bounds do not match dimension of params.")
    if len(upper_flat) != n_params:
        raise InvalidBoundsError("upper_bounds do not match dimension of params.")

    lower_flat[np.isnan(lower_flat)] = -np.inf
    upper_flat[np.isnan(upper_flat)] = np.inf

    if add_soft_bounds:
        lower_flat_soft = _update_bounds_and_flatten(
            nan_tree, bounds.soft_lower, kind="soft_lower_bound"
        )
        lower_flat_soft[np.isnan(lower_flat_soft)] = -np.inf
        lower_flat = np.maximum(lower_flat, lower_flat_soft)

        upper_flat_soft = _update_bounds_and_flatten(
            nan_tree, bounds.soft_upper, kind="soft_upper_bound"
        )
        upper_flat_soft[np.isnan(upper_flat_soft)] = np.inf
        upper_flat = np.minimum(upper_flat, upper_flat_soft)

    if (lower_flat > upper_flat).any():
        msg = "Invalid bounds. Some lower bounds are larger than upper bounds."
        raise InvalidBoundsError(msg)

    return lower_flat, upper_flat


def _update_bounds_and_flatten(
    nan_tree: PyTree, bounds: PyTree, kind: str
) -> NDArray[np.float64]:
    """Flatten bounds array and update it with bounds from params.

    Args:
        nan_tree: Pytree with the same structure as params, filled with nans.
        bounds: The candidate bounds to be updated and flattened.
        kind: One of "lower_bound", "upper_bound", "soft_lower_bound",
            "soft_upper_bound".

    Returns:
        np.ndarray: The updated and flattened bounds.

    """
    registry = get_registry(extended=True, data_col=kind)
    flat_nan_tree = tree_leaves(nan_tree, registry=registry)

    if bounds is not None:
        registry = get_registry(extended=True)
        flat_bounds = tree_leaves(bounds, registry=registry)

        seperator = 10 * "$"
        params_names = leaf_names(nan_tree, registry=registry, separator=seperator)
        bounds_names = leaf_names(bounds, registry=registry, separator=seperator)

        flat_nan_dict = dict(zip(params_names, flat_nan_tree, strict=False))

        invalid = {"names": [], "bounds": []}  # type: ignore
        for bounds_name, bounds_leaf in zip(bounds_names, flat_bounds, strict=False):
            # if a bounds leaf is None we treat it as saying the the corresponding
            # subtree of params has no bounds.
            if bounds_leaf is not None:
                if bounds_name in flat_nan_dict:
                    flat_nan_dict[bounds_name] = bounds_leaf
                else:
                    invalid["names"].append(bounds_name)
                    invalid["bounds"].append(bounds_leaf)

        if invalid["bounds"]:
            msg = (
                f"{kind} could not be matched to params pytree. The bounds "
                f"{invalid['bounds']} with names {invalid['names']} are not part of "
                "params."
            )
            raise InvalidBoundsError(msg)

        flat_nan_tree = list(flat_nan_dict.values())

    updated = np.array(flat_nan_tree, dtype=np.float64)
    return updated


def _is_fast_path(params: PyTree, bounds: Bounds, add_soft_bounds: bool) -> bool:
    out = True
    if add_soft_bounds:
        out = False

    if not _is_1d_array(params):
        out = False

    for bound in bounds.lower, bounds.upper:
        if not (_is_1d_array(bound) or bound is None):
            out = False
    return out


def _is_1d_array(candidate: Any) -> bool:
    return isinstance(candidate, np.ndarray) and candidate.ndim == 1


def _get_fast_path_bounds(
    params: PyTree, bounds: Bounds
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    if bounds.lower is None:
        # faster than np.full
        lower_bounds = np.array([-np.inf] * len(params))
    else:
        lower_bounds = bounds.lower.astype(float)

    if bounds.upper is None:
        # faster than np.full
        upper_bounds = np.array([np.inf] * len(params))
    else:
        upper_bounds = bounds.upper.astype(float)

    if (lower_bounds > upper_bounds).any():
        msg = "Invalid bounds. Some lower bounds are larger than upper bounds."
        raise InvalidBoundsError(msg)

    return lower_bounds, upper_bounds



import logging
import warnings
from dataclasses import replace
from functools import wraps
from pathlib import Path
from typing import Any, Callable, ParamSpec, cast

from optimagic import mark
from optimagic.constraints import Constraint, InvalidConstraintError
from optimagic.logging.logger import (
    LogOptions,
    SQLiteLogOptions,
)
from optimagic.optimization.fun_value import (
    LeastSquaresFunctionValue,
    LikelihoodFunctionValue,
    ScalarFunctionValue,
)
from optimagic.parameters.bounds import Bounds
from optimagic.typing import AggregationLevel

_logger = logging.getLogger(__name__)


def throw_criterion_future_warning():
    msg = (
        "To align optimagic with scipy.optimize, the `criterion` argument has been "
        "renamed to `fun`. Please use `fun` instead of `criterion`. Using `criterion` "
        " will become an error in optimagic version 0.6.0 and later."
    )
    warnings.warn(msg, FutureWarning)


def throw_criterion_kwargs_future_warning():
    msg = (
        "To align optimagic with scipy.optimize, the `criterion_kwargs` argument has "
        "been renamed to `fun_kwargs`. Please use `fun_kwargs` instead of "
        "`criterion_kwargs`. Using `criterion_kwargs` will become an error in "
        "optimagic version 0.6.0 and later."
    )
    warnings.warn(msg, FutureWarning)


def throw_derivative_future_warning():
    msg = (
        "To align optimagic with scipy.optimize, the `derivative` argument has been "
        "renamed to `jac`. Please use `jac` instead of `derivative`. Using `derivative`"
        " will become an error in optimagic version 0.6.0 and later. For more details "
        "see the documentation: "
        "https://optimagic.readthedocs.io/en/latest/how_to/how_to_derivatives.html"
    )
    warnings.warn(msg, FutureWarning)


def throw_derivative_kwargs_future_warning():
    msg = (
        "To align optimagic with scipy.optimize, the `derivative_kwargs` argument has "
        "been renamed to `jac_kwargs`. Please use `jac_kwargs` instead of "
        "`derivative_kwargs`. Using `derivative_kwargs` will become an error in "
        "optimagic version 0.6.0 and later. For more details see the documentation: "
        "https://optimagic.readthedocs.io/en/latest/how_to/how_to_derivatives.html"
    )
    warnings.warn(msg, FutureWarning)


def throw_criterion_and_derivative_future_warning():
    msg = (
        "To align optimagic with scipy.optimize, the `criterion_and_derivative` "
        "argument has been renamed to `fun_and_jac`. Please use `fun_and_jac` "
        "instead of `criterion_and_derivative`. Using `criterion_and_derivative` "
        "will become an error in optimagic version 0.6.0 and later. For more details "
        "see the documentation: "
        "https://optimagic.readthedocs.io/en/latest/how_to/how_to_derivatives.html"
    )
    warnings.warn(msg, FutureWarning)


def throw_criterion_and_derivative_kwargs_future_warning():
    msg = (
        "To align optimagic with scipy.optimize, the `criterion_and_derivative_kwargs` "
        "argument has been renamed to `fun_and_jac_kwargs`. Please use "
        "`fun_and_jac_kwargs` instead of `criterion_and_derivative_kwargs`. Using "
        "`criterion_and_derivative_kwargs` will become an error in optimagic version "
        "0.6.0 and later. For more details see the documentation: "
        "https://optimagic.readthedocs.io/en/latest/how_to/how_to_derivatives.html"
    )
    warnings.warn(msg, FutureWarning)


def throw_scaling_options_future_warning():
    msg = (
        "Specifying scaling options via the argument `scaling_options` is deprecated "
        "and will be removed in optimagic version 0.6.0 and later. You can pass these "
        "options directly to the `scaling` argument instead."
    )
    warnings.warn(msg, FutureWarning)


def throw_multistart_options_future_warning():
    msg = (
        "Specifying multistart options via the argument `multistart_options` is "
        "deprecated and will be removed in optimagic version 0.6.0 and later. You can "
        "pass these options directly to the `multistart` argument instead. For more "
        "details see the documentation: "
        "https://optimagic.readthedocs.io/en/latest/how_to/how_to_multistart.html"
    )
    warnings.warn(msg, FutureWarning)


def throw_derivatives_step_ratio_future_warning():
    msg = (
        "The `step_ratio` argument is deprecated and will be removed alongside "
        "Richardson extrapolation in optimagic version 0.6.0."
    )
    warnings.warn(msg, FutureWarning)


def throw_derivatives_n_steps_future_warning():
    msg = (
        "The `n_steps` argument is deprecated and will be removed alongside "
        "Richardson extrapolation in optimagic version 0.6.0."
    )
    warnings.warn(msg, FutureWarning)


def throw_derivatives_return_info_future_warning():
    msg = (
        "The `return_info` argument is deprecated and will be removed alongside "
        "Richardson extrapolation in optimagic version 0.6.0."
    )
    warnings.warn(msg, FutureWarning)


def throw_derivatives_return_func_value_future_warning():
    msg = (
        "The `return_func_value` argument is deprecated and will be removed in "
        "optimagic version 0.6.0."
    )
    warnings.warn(msg, FutureWarning)


def throw_numdiff_result_func_evals_future_warning():
    msg = (
        "The `func_evals` attribute is deprecated and will be removed in optimagic "
        "version 0.6.0."
    )
    warnings.warn(msg, FutureWarning)


def throw_numdiff_result_derivative_candidates_future_warning():
    msg = (
        "The `derivative_candidates` attribute is deprecated and will be removed in "
        "optimagic version 0.6.0."
    )
    warnings.warn(msg, FutureWarning)


def throw_numdiff_options_deprecated_in_estimate_ml_future_warning():
    msg = (
        "The argument `numdiff_options` is deprecated for `estimate_ml` and will be "
        "removed in estimagic version 0.6.0. Please use the `jacobian_numdiff_options` "
        "and `hessian_numdiff_options` arguments instead to specify the options for "
        "the first and second numerical derivative estimation."
    )
    warnings.warn(msg, FutureWarning)


def throw_numdiff_options_deprecated_in_estimate_msm_future_warning():
    msg = (
        "The argument `numdiff_options` is deprecated for `estimate_msm` and will be "
        "removed in estimagic version 0.6.0. Please use the `jacobian_numdiff_options` "
        "argument instead."
    )
    warnings.warn(msg, FutureWarning)


def throw_dict_access_future_warning(attribute, obj_name):
    msg = (
        f"The dictionary access for '{attribute}' is deprecated and will be removed "
        "in optimagic version 0.6.0. Please use the new attribute access instead: "
        f"`{obj_name}.{attribute}`."
    )
    warnings.warn(msg, FutureWarning)


def replace_and_warn_about_deprecated_algo_options(algo_options):
    if not isinstance(algo_options, dict):
        return algo_options

    algo_options = {k.replace(".", "_"): v for k, v in algo_options.items()}

    replacements = {
        "stopping_max_criterion_evaluations": "stopping_maxfun",
        "stopping_max_iterations": "stopping_maxiter",
        "convergence_absolute_criterion_tolerance": "convergence_ftol_abs",
        "convergence_relative_criterion_tolerance": "convergence_ftol_rel",
        "convergence_scaled_criterion_tolerance": "convergence_ftol_scaled",
        "convergence_absolute_params_tolerance": "convergence_xtol_abs",
        "convergence_relative_params_tolerance": "convergence_xtol_rel",
        "convergence_absolute_gradient_tolerance": "convergence_gtol_abs",
        "convergence_relative_gradient_tolerance": "convergence_gtol_rel",
        "convergence_scaled_gradient_tolerance": "convergence_gtol_scaled",
    }

    present = sorted(set(algo_options) & set(replacements))
    if present:
        msg = (
            "The following keys in `algo_options` are deprecated and will be removed "
            "in optimagic version 0.6.0 and later. Please replace them as follows:\n"
        )
        for k in present:
            msg += f"  {k} -> {replacements[k]}\n"

        warnings.warn(msg, FutureWarning)

    out = {k: v for k, v in algo_options.items() if k not in present}
    for k in present:
        out[replacements[k]] = algo_options[k]

    return out


def replace_and_warn_about_deprecated_bounds(
    lower_bounds,
    upper_bounds,
    bounds,
    soft_lower_bounds=None,
    soft_upper_bounds=None,
):
    old_bounds = {
        "lower": lower_bounds,
        "upper": upper_bounds,
        "soft_lower": soft_lower_bounds,
        "soft_upper": soft_upper_bounds,
    }

    old_present = [k for k, v in old_bounds.items() if v is not None]

    if old_present:
        substring = ", ".join(f"{b}_bound" for b in old_present)
        substring = substring.replace(", ", ", and ", -1)
        msg = (
            f"Specifying bounds via the arguments {substring} is "
            "deprecated and will be removed in optimagic version 0.6.0 and later. "
            "Please use the `bounds` argument instead."
        )
        warnings.warn(msg, FutureWarning)

    if bounds is None and old_present:
        bounds = Bounds(**old_bounds)

    return bounds


def convert_dict_to_function_value(candidate):
    """Convert the deprecated dictionary output to a suitable FunctionValue object.

    No warning is raised here because this function will be called repeatedly!

    """
    special_keys = ["value", "contributions", "root_contributions"]

    if is_dict_output(candidate):
        info = {k: v for k, v in candidate.items() if k not in special_keys}
        if "root_contributions" in candidate:
            out = LeastSquaresFunctionValue(candidate["root_contributions"], info)
        elif "contributions" in candidate:
            out = LikelihoodFunctionValue(candidate["contributions"], info)
        else:
            out = ScalarFunctionValue(candidate["value"], info)
    else:
        out = candidate

    return out


def is_dict_output(candidate):
    """Check if the output is a dictionary with special keys."""
    special_keys = ["value", "contributions", "root_contributions"]
    return isinstance(candidate, dict) and any(k in candidate for k in special_keys)


def throw_dict_output_warning():
    msg = (
        "Returning a dictionary with the special keys 'value', 'contributions', or "
        "'root_contributions' is deprecated and will be removed in optimagic version "
        "0.6.0 and later. Please use the optimagic.mark.scalar, optimagic.mark."
        "least_squares, or optimagic.mark.likelihood decorators to indicate the type "
        "of problem you are solving. Use optimagic.FunctionValue objects to return "
        "additional information for the logging. Please see the documentation for more "
        "details: https://optimagic.readthedocs.io/en/latest/how_to/how_to_criterion_function.html"
    )
    warnings.warn(msg, FutureWarning)


def infer_problem_type_from_dict_output(output):
    if "root_contributions" in output:
        out = AggregationLevel.LEAST_SQUARES
    elif "contributions" in output:
        out = AggregationLevel.LIKELIHOOD
    else:
        out = AggregationLevel.SCALAR
    return out


P = ParamSpec("P")


def replace_dict_output(func: Callable[P, Any]) -> Callable[P, Any]:
    """Replace the deprecated dictionary output by a suitable FunctionValue.

    This has no effect if the function does not return a dictionary with at least one of
    the special keys "value", "contributions" or "root_contributions" or a tuple where
    the first entry is such a dictionary.

    This decorator does not add a warning because the function will be evaluated many
    times and the warning would pop up too often.

    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
        raw = func(*args, **kwargs)
        # fun and jac case
        if isinstance(raw, tuple):
            out = (convert_dict_to_function_value(raw[0]), raw[1])
        # fun case
        else:
            out = convert_dict_to_function_value(raw)
        return out

    return wrapper


def throw_key_warning_in_derivatives():
    msg = (
        "The `key` argument in first_derivative and second_derivative is deprecated "
        "and will be removed in optimagic version 0.6.0 and later. Please use the "
        "`unpacker` argument instead. While `key` was a string, `unpacker` is a "
        "callable that takes the output of `func` and returns the desired output that "
        "is then differentiated."
    )
    warnings.warn(msg, FutureWarning)


def throw_dict_constraints_future_warning_if_required(
    constraints: list[dict[str, Any]] | dict[str, Any],
) -> None:
    replacements = {
        "fixed": "optimagic.FixedConstraint",
        "increasing": "optimagic.IncreasingConstraint",
        "decreasing": "optimagic.DecreasingConstraint",
        "equality": "optimagic.EqualityConstraint",
        "probability": "optimagic.ProbabilityConstraint",
        "pairwise_equality": "optimagic.PairwiseEqualityConstraint",
        "covariance": "optimagic.FlatCovConstraint",
        "sdcorr": "optimagic.FlatSDCorrConstraint",
        "linear": "optimagic.LinearConstraint",
        "nonlinear": "optimagic.NonlinearConstraint",
    }

    if not isinstance(constraints, list):
        constraints = [constraints]

    types = [
        constraint.get("type", None) if isinstance(constraint, dict) else None
        for constraint in constraints
    ]
    types = list(set(types) - {None})

    if types:
        msg = (
            "Specifying constraints as a dictionary is deprecated and will be removed "
            "in optimagic version 0.6.0. Please replace them using the new optimagic "
            "constraint objects:\n"
        )
        for t in types:
            msg += f"  {{'type': '{t}', ...}} -> {replacements[t]}(...)\n"
        msg += (
            "\nFor more details see the documentation: "
            "https://optimagic.readthedocs.io/en/latest/how_to/how_to_constraints.html"
        )

        warnings.warn(msg, FutureWarning)


def replace_and_warn_about_deprecated_multistart_options(options):
    """Replace deprecated multistart options and warn about them.

    Args:
        options (MultistartOptions): The multistart options to replace.

    Returns:
        MultistartOptions: The replaced multistart options.

    """
    replacements = {}

    if options.share_optimization is not None:
        msg = (
            "The `share_optimization` option is deprecated and will be removed in "
            "version 0.6.0. Use `stopping_maxopt` instead to specify the number of "
            "optimizations directly. For more details see the documentation: "
            "https://optimagic.readthedocs.io/en/latest/how_to/how_to_multistart.html"
        )
        warnings.warn(msg, FutureWarning)

    if options.convergence_relative_params_tolerance is not None:
        msg = (
            "The `convergence_relative_params_tolerance` option is deprecated and will "
            "be removed in version 0.6.0. Use `convergence_xtol_rel` instead. For more "
            "details see the documentation: "
            "https://optimagic.readthedocs.io/en/latest/how_to/how_to_multistart.html"
        )
        warnings.warn(msg, FutureWarning)
        if options.convergence_xtol_rel is None:
            replacements["convergence_xtol_rel"] = (
                options.convergence_relative_params_tolerance
            )

    if options.optimization_error_handling is not None:
        msg = (
            "The `optimization_error_handling` option is deprecated and will be "
            "removed in version 0.6.0. Setting this attribute also sets the error "
            "handling for exploration. Use the new `error_handling` option to set the "
            "error handling for both optimization and exploration. For more details "
            "see the documentation: "
            "https://optimagic.readthedocs.io/en/latest/how_to/how_to_multistart.html"
        )
        warnings.warn(msg, FutureWarning)
        if options.error_handling is None:
            replacements["error_handling"] = options.optimization_error_handling

    if options.exploration_error_handling is not None:
        msg = (
            "The `exploration_error_handling` option is deprecated and will be "
            "removed in version 0.6.0. Setting this attribute also sets the error "
            "handling for exploration. Use the new `error_handling` option to set the "
            "error handling for both optimization and exploration. For more details "
            "see the documentation: "
            "https://optimagic.readthedocs.io/en/latest/how_to/how_to_multistart.html"
        )
        warnings.warn(msg, FutureWarning)
        if options.error_handling is None:
            replacements["error_handling"] = options.exploration_error_handling

    return replace(options, **replacements)


def replace_and_warn_about_deprecated_base_steps(
    step_size,
    base_steps,
):
    if base_steps is not None:
        msg = (
            "The `base_steps` argument is deprecated and will be removed alongside "
            "Richardson extrapolation in optimagic version 0.6.0. To specify the "
            "step size use the `step_size` argument instead. For more details see the "
            "documentation: "
            "https://optimagic.readthedocs.io/en/latest/how_to/how_to_step_size.html"
        )
        warnings.warn(msg, FutureWarning)

        if step_size is None:
            step_size = base_steps

    return step_size


def replace_and_warn_about_deprecated_derivatives(candidate, name):
    msg = (
        f"Specifying a dictionary of {name} functions is deprecated and will be "
        "removed in optimagic version 0.6.0. Please specify a single function that "
        "returns the correct derivative for your optimizer or a list of functions that "
        "are decorated with the `mark.scalar`, `mark.likelihood` or "
        "`mark.least_squares` decorators. For more details see the documentation: "
        "https://optimagic.readthedocs.io/en/latest/how_to/how_to_derivatives.html"
    )
    warnings.warn(msg, FutureWarning)

    key_to_marker = {
        "value": mark.scalar,
        "contributions": mark.likelihood,
        "root_contributions": mark.least_squares,
    }

    out = []
    for key, func in candidate.items():
        if key in key_to_marker:
            out.append(key_to_marker[key](func))

    return out


def handle_log_options_throw_deprecated_warning(
    log_options: dict[str, Any], logger: str | Path | LogOptions | None
) -> str | Path | LogOptions | None:
    msg = (
        "Usage of the parameter log_options is deprecated "
        "and will be removed in a future version. "
        "Provide a LogOptions instance for the parameter `logging`, if you need to "
        "configure the logging."
    )
    warnings.warn(msg, FutureWarning)
    logging_is_path_or_string = isinstance(logger, str) or isinstance(logger, Path)
    log_options_is_dict = isinstance(log_options, dict)
    compatible_keys = {"fast_logging", "if_table_exists", "if_database_exists"}
    log_options_is_compatible = set(log_options.keys()).issubset(compatible_keys)

    if logging_is_path_or_string:
        if log_options_is_dict and log_options_is_compatible:
            warnings.warn(
                f"\nUsing {log_options=} to create an instance of SQLiteLogOptions. "
                f"This mechanism will be removed in the future.",
                FutureWarning,
            )
            if "if_table_exists" in log_options:
                warnings.warn(
                    "Found 'if_table_exists' in options dictionary. "
                    "This option is deprecated and setting it has no effect.",
                    FutureWarning,
                )
                log_options = {
                    k: v for k, v in log_options.items() if k != "if_table_exists"
                }
            return SQLiteLogOptions(cast(str | Path, logger), **log_options)
        elif not log_options_is_compatible:
            raise ValueError(
                f"Found string or path for logger argument, but parameter"
                f" {log_options=} is not compatible to {compatible_keys=}."
                f"Explicitly create a Logger instance for configuration."
            )

    return logger


def pre_process_constraints(
    constraints: list[Constraint | dict[str, Any]] | Constraint | dict[str, Any] | None,
) -> list[dict[str, Any]]:
    """Convert all ways of specifying constraints to a list of dictionaries.

    For the optimagic release 0.5.0 we only implemented the new constraint API, but have
    not overhauled the internal representation of constraints yet. As a result, we
    convert all ways of specifying constraints, and in particular the new interface, to
    the old format, that is, a list of dictionaries.

    Once we have refactor the internal representation of constraints, we will be able to
    go the other way, and convert all formats to the new one.

    """
    if constraints is None:
        return []

    if isinstance(constraints, dict | Constraint):
        constraints = [constraints]

    if isinstance(constraints, list):
        out = []
        invalid_types: list[type] = []
        for constr in constraints:
            if isinstance(constr, Constraint):
                out.append(constr._to_dict())
            elif isinstance(constr, dict):
                out.append(constr)
            else:
                invalid_types.append(type(constr))

        if invalid_types:
            msg = (
                f"Invalid constraint types: {set(invalid_types)}. Must be a constraint "
                "object imported from `optimagic`."
            )
            raise InvalidConstraintError(msg)

    else:
        msg = (
            f"Invalid constraint type: {type(constraints)}. Must be a constraint "
            "object or list thereof imported from `optimagic`. For more details see "
            "the documentation: "
            "https://optimagic.readthedocs.io/en/latest/how_to/how_to_constraints.html"
        )
        raise InvalidConstraintError(msg)

    return out



from __future__ import annotations

import warnings

from optimagic import constraints, mark, utilities
from optimagic.algorithms import algos
from optimagic.benchmarking.benchmark_reports import (
    convergence_report,
    rank_report,
    traceback_report,
)
from optimagic.benchmarking.get_benchmark_problems import get_benchmark_problems
from optimagic.benchmarking.run_benchmark import run_benchmark
from optimagic.constraints import (
    DecreasingConstraint,
    EqualityConstraint,
    FixedConstraint,
    FlatCovConstraint,
    FlatSDCorrConstraint,
    IncreasingConstraint,
    LinearConstraint,
    NonlinearConstraint,
    PairwiseEqualityConstraint,
    ProbabilityConstraint,
)
from optimagic.differentiation.derivatives import first_derivative, second_derivative
from optimagic.differentiation.numdiff_options import NumdiffOptions
from optimagic.logging import (
    ExistenceStrategy as ExistenceStrategy,
)
from optimagic.logging import (
    SQLiteLogOptions as SQLiteLogOptions,
)
from optimagic.logging import (
    SQLiteLogReader as SQLiteLogReader,
)
from optimagic.logging.read_log import OptimizeLogReader
from optimagic.optimization.fun_value import (
    FunctionValue,
    LeastSquaresFunctionValue,
    LikelihoodFunctionValue,
    ScalarFunctionValue,
)
from optimagic.optimization.history import History
from optimagic.optimization.multistart_options import MultistartOptions
from optimagic.optimization.optimize import maximize, minimize
from optimagic.optimization.optimize_result import OptimizeResult
from optimagic.parameters.bounds import Bounds
from optimagic.parameters.constraint_tools import check_constraints, count_free_params
from optimagic.parameters.scaling import ScalingOptions
from optimagic.visualization.convergence_plot import convergence_plot
from optimagic.visualization.history_plots import criterion_plot, params_plot
from optimagic.visualization.profile_plot import profile_plot
from optimagic.visualization.slice_plot import slice_plot

try:
    from ._version import version as __version__
except ImportError:
    # broken installation, we don't even try unknown only works because we do poor mans
    # version compare
    __version__ = "unknown"

installation_warning_message = (
    "estimagic has been renamed to optimagic, but you have estimagic installed. Please "
    "uninstall estimagic and install optimagic instead. Don't worry, your estimagic "
    "imports will still work if you install optimagic, and simple warnings will help "
    "you to adjust them for future releases.\n\n"
    "To make these changes using pip, run:\n"
    "-------------------------------------\n"
    "$ pip uninstall estimagic\n"
    "$ pip install optimagic\n\n"
    "For conda users, use:\n"
    "---------------------\n"
    "$ conda remove estimagic\n"
    "$ conda install -c conda-forge optimagic\n\n"
    "For more details see the documentation: "
    "https://optimagic.readthedocs.io/en/latest/installation.html"
)
warnings.warn(installation_warning_message, FutureWarning)


__all__ = [
    "maximize",
    "minimize",
    "utilities",
    "first_derivative",
    "second_derivative",
    "run_benchmark",
    "get_benchmark_problems",
    "profile_plot",
    "convergence_plot",
    "convergence_report",
    "rank_report",
    "traceback_report",
    "slice_plot",
    "criterion_plot",
    "params_plot",
    "count_free_params",
    "check_constraints",
    "OptimizeLogReader",
    "OptimizeResult",
    "Bounds",
    "mark",
    "ScalingOptions",
    "MultistartOptions",
    "NumdiffOptions",
    "FunctionValue",
    "LeastSquaresFunctionValue",
    "ScalarFunctionValue",
    "LikelihoodFunctionValue",
    "constraints",
    "FlatCovConstraint",
    "FlatSDCorrConstraint",
    "IncreasingConstraint",
    "DecreasingConstraint",
    "FixedConstraint",
    "NonlinearConstraint",
    "LinearConstraint",
    "ProbabilityConstraint",
    "PairwiseEqualityConstraint",
    "EqualityConstraint",
    "History",
    "__version__",
    "algos",
]



from dataclasses import dataclass
from typing import Literal, TypedDict

from typing_extensions import NotRequired

from optimagic.exceptions import InvalidScalingError


@dataclass(frozen=True)
class ScalingOptions:
    """Scaling options in optimization problems.

    Attributes:
        method: The method used for scaling. Can be "start_values" or "bounds". Default
            is "start_values".
        clipping_value: The minimum value to which elements are clipped to avoid
            division by zero. Must be a positive number. Default is 0.1.
        magnitude: A factor by which the scaled parameters are multiplied to adjust
            their magnitude. Must be a positive number. Default is 1.0.

    Raises:
        InvalidScalingError: If scaling options cannot be processed, e.g. because they
            do not have the correct type.

    """

    method: Literal["start_values", "bounds"] = "start_values"
    clipping_value: float = 0.1
    magnitude: float = 1.0

    def __post_init__(self) -> None:
        _validate_attribute_types_and_values(self)


class ScalingOptionsDict(TypedDict):
    method: NotRequired[Literal["start_values", "bounds"]]
    clipping_value: NotRequired[float]
    magnitude: NotRequired[float]


def pre_process_scaling(
    scaling: bool | ScalingOptions | ScalingOptionsDict | None,
) -> ScalingOptions | None:
    """Convert all valid types of scaling options to optimagic.ScalingOptions.

    This just harmonizes multiple ways of specifying scaling options into a single
    format. It performs runtime type checks, but it does not check whether scaling
    options are consistent with other option choices.

    Args:
        scaling: The user provided scaling options.

    Returns:
        The scaling options in the optimagic format.

    Raises:
        InvalidScalingError: If scaling options cannot be processed, e.g. because they
            do not have the correct type.

    """
    if isinstance(scaling, bool):
        scaling = ScalingOptions() if scaling else None
    elif isinstance(scaling, ScalingOptions) or scaling is None:
        pass
    else:
        try:
            scaling = ScalingOptions(**scaling)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            if isinstance(e, InvalidScalingError):
                raise e
            raise InvalidScalingError(
                f"Invalid scaling options of type: {type(scaling)}. Scaling options "
                "must be of type optimagic.ScalingOptions, a dictionary with a subset "
                "of the keys {'method', 'clipping_value', 'magnitude'}, None, or a "
                "boolean."
            ) from e

    return scaling


def _validate_attribute_types_and_values(options: ScalingOptions) -> None:
    if options.method not in ("start_values", "bounds"):
        raise InvalidScalingError(
            f"Invalid scaling method: {options.method}. Valid methods are "
            "'start_values' and 'bounds'."
        )

    if (
        not isinstance(options.clipping_value, int | float)
        or options.clipping_value <= 0
    ):
        raise InvalidScalingError(
            f"Invalid clipping value: {options.clipping_value}. Clipping value "
            "must be a positive number."
        )

    if not isinstance(options.magnitude, int | float) or options.magnitude <= 0:
        raise InvalidScalingError(
            f"Invalid scaling magnitude: {options.magnitude}. Scaling magnitude "
            "must be a positive number."
        )



import warnings
from dataclasses import dataclass
from typing import Any

import numpy as np

from optimagic.typing import EvalTask, PyTree


@dataclass(frozen=True)
class HistoryEntry:
    params: PyTree
    fun: float | None
    time: float
    task: EvalTask


class History:
    # TODO: add counters for the relevant evaluations
    def __init__(self) -> None:
        self._params: list[PyTree] = []
        self._fun: list[float | None] = []
        self._time: list[float] = []
        self._batches: list[int] = []
        self._task: list[EvalTask] = []

    def add_entry(self, entry: HistoryEntry, batch_id: int | None = None) -> None:
        if batch_id is None:
            batch_id = self._get_next_batch_id()
        self._params.append(entry.params)
        self._fun.append(entry.fun)
        self._time.append(entry.time)
        self._batches.append(batch_id)
        self._task.append(entry.task)

    def add_batch(
        self, batch: list[HistoryEntry], batch_size: int | None = None
    ) -> None:
        # The naming is complicated here:
        # batch refers to the entries to be added to the history in one go
        # batch_size is a property of a parallelizing algorithm that influences how
        # the batch_ids are assigned. It is not the same as the length of the batch.
        if batch_size is None:
            batch_size = len(batch)

        start = self._get_next_batch_id()
        n_batches = int(np.ceil(len(batch) / batch_size))
        ids = np.repeat(np.arange(start, start + n_batches), batch_size)[: len(batch)]

        for entry, id in zip(batch, ids, strict=False):
            self.add_entry(entry, id)

    @property
    def params(self) -> list[PyTree]:
        return self._params

    @property
    def fun(self) -> list[float | None]:
        return self._fun

    @property
    def time(self) -> list[float]:
        arr = np.array(self._time)
        return (arr - arr[0]).tolist()

    @property
    def batches(self) -> list[int]:
        return self._batches

    @property
    def task(self) -> list[EvalTask]:
        return self._task

    def _get_next_batch_id(self) -> int:
        if not self._batches:
            batch = 0
        else:
            batch = self._batches[-1] + 1
        return batch

    # ==================================================================================
    # Add deprecated dict access
    # ==================================================================================

    @property
    def criterion(self) -> list[float | None]:
        msg = "The attribute `criterion` of History is deprecated. Use `fun` instead."
        warnings.warn(msg, FutureWarning)
        return self.fun

    @property
    def runtime(self) -> list[float]:
        msg = "The attribute `runtime` of History is deprecated. Use `time` instead."
        warnings.warn(msg, FutureWarning)
        return self.time

    def __getitem__(self, key: str) -> Any:
        msg = "dict-like access to History is deprecated. Use attribute access instead."
        warnings.warn(msg, FutureWarning)
        return getattr(self, key)



import functools

from optimagic.exceptions import InvalidFunctionError
from optimagic.utilities import propose_alternatives


def map_method_to_algorithm(method):
    implemented = {
        "Nelder-Mead": "scipy_neldermead",
        "Powell": "scipy_powell",
        "CG": "scipy_conjugate_gradient",
        "BFGS": "scipy_bfgs",
        "Newton-CG": "scipy_newton_cg",
        "L-BFGS-B": "scipy_lbfgsb",
        "TNC": "scipy_truncated_newton",
        "COBYLA": "scipy_cobyla",
        "SLSQP": "scipy_slsqp",
        "trust-constr": "scipy_trust_constr",
    }

    not_implemented = {
        "dogleg": "scipy_dogleg",
        "trust-ncg": "scipy_trust_ncg",
        "trust-exact": "scipy_trust_exact",
        "trust-krylov": "scipy_trust_krylov",
        "COBYQA": "scipy_cobyqa",
    }

    if method in implemented:
        algo = implemented[method]
    elif method in not_implemented:
        msg = (
            f"The method {method} is not yet wrapped in optimagic. Create an issue on "
            "https://github.com/optimagic-dev/optimagic/ if you have urgent need "
            "for this method."
        )
        raise NotImplementedError(msg)
    else:
        alt = propose_alternatives(method, list(implemented) + list(not_implemented))
        msg = (
            "method is an alias for algorithm to select the scipy optimizers under "
            f"their original name. {method} is not a valid scipy algorithm name. "
            f"Did you mean {alt}?"
        )
        raise ValueError(msg)
    return algo


def split_fun_and_jac(fun_and_jac, target="fun"):
    index = 0 if target == "fun" else 1

    @functools.wraps(fun_and_jac)
    def fun(*args, **kwargs):
        raw = fun_and_jac(*args, **kwargs)
        try:
            out = raw[index]
        except TypeError as e:
            msg = (
                "If you set `jac=True`, `fun` needs to return a tuple where the first "
                "entry is the value of your objective function and the second entry "
                "is its derivative."
            )
            raise InvalidFunctionError(msg) from e
        return out

    return fun



"""Import common objective functions in several optimagic compatible versions.

All implemented functions accept arbitrary pytrees as parameters. If possible they are
implemented as scalar and least-squares versions.

"""

from __future__ import annotations

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from pybaum import tree_just_flatten, tree_unflatten

from optimagic import mark
from optimagic.optimization.fun_value import (
    FunctionValue,
)
from optimagic.parameters.block_trees import matrix_to_block_tree
from optimagic.parameters.tree_registry import get_registry
from optimagic.typing import PyTree

REGISTRY = get_registry(extended=True)


@mark.scalar
def trid_scalar(params: PyTree) -> float:
    """Implement Trid function: https://www.sfu.ca/~ssurjano/trid.html."""
    x = _get_x(params)
    return ((x - 1) ** 2).sum() - (x[1:] * x[:-1]).sum()


@mark.scalar
def trid_gradient(params: PyTree) -> PyTree:
    """Calculate gradient of trid function."""
    x = _get_x(params)
    l1 = np.insert(x, 0, 0)
    l1 = np.delete(l1, [-1])
    l2 = np.append(x, 0)
    l2 = np.delete(l2, [0])
    flat = 2 * (x - 1) - l1 - l2
    return _unflatten_gradient(flat, params)


@mark.scalar
def trid_fun_and_gradient(params: PyTree) -> tuple[float, PyTree]:
    """Implement Trid function and calculate gradient."""
    val = trid_scalar(params)
    grad = trid_gradient(params)
    return val, grad


@mark.scalar
def rhe_scalar(params: PyTree) -> float:
    """Implement Rotated Hyper Ellipsoid function.

    Function description: https://www.sfu.ca/~ssurjano/rothyp.html.

    """
    return (rhe_ls(params) ** 2).sum()


@mark.scalar
def rhe_gradient(params: PyTree) -> PyTree:
    """Calculate gradient of rotated_hyper_ellipsoid function."""
    x = _get_x(params)
    flat = np.arange(2 * len(x), 0, -2) * x
    return _unflatten_gradient(flat, params)


@mark.scalar
def rhe_fun_and_gradient(params: PyTree) -> tuple[float, PyTree]:
    """Implement Rotated Hyper Ellipsoid function and calculate gradient."""
    val = rhe_scalar(params)
    grad = rhe_gradient(params)
    return val, grad


@mark.least_squares
def rhe_ls(params: PyTree) -> NDArray[np.float64]:
    """Compute least-squares version of the Rotated Hyper Ellipsoid function."""
    x = _get_x(params)
    dim = len(params)
    out = np.zeros(dim)
    for i in range(dim):
        out[i] = np.sqrt((x[: i + 1] ** 2).sum())
    return out


@mark.least_squares
def rhe_function_value(params: PyTree) -> FunctionValue:
    """FunctionValue version of Rotated Hyper Ellipsoid function."""
    contribs = rhe_ls(params)
    out = FunctionValue(contribs)
    return out


@mark.scalar
def rosenbrock_scalar(params: PyTree) -> float:
    """Rosenbrock function: https://www.sfu.ca/~ssurjano/rosen.html."""
    return (rosenbrock_ls(params) ** 2).sum()


@mark.scalar
def rosenbrock_gradient(params: PyTree) -> PyTree:
    """Calculate gradient of rosenbrock function."""
    x = _get_x(params)
    l1 = np.delete(x, [-1])
    l1 = np.append(l1, 0)
    l2 = np.insert(x, 0, 0)
    l2 = np.delete(l2, [1])
    l3 = np.insert(x, 0, 0)
    l3 = np.delete(l3, [-1])
    l4 = np.delete(x, [0])
    l4 = np.append(l4, 0)
    l5 = np.full((len(x) - 1), 2)
    l5 = np.append(l5, 0)
    flat = 100 * (4 * (l1**3) + 2 * l2 - 2 * (l3**2) - 4 * (l4 * x)) + 2 * l1 - l5
    return _unflatten_gradient(flat, params)


@mark.scalar
def rosenbrock_fun_and_gradient(params: PyTree) -> tuple[float, PyTree]:
    """Implement rosenbrock function and calculate gradient."""
    return rosenbrock_scalar(params), rosenbrock_gradient(params)


@mark.least_squares
def rosenbrock_ls(params: PyTree) -> NDArray[np.float64]:
    """Least-squares version of the rosenbrock function."""
    x = _get_x(params)
    dim = len(params)
    out = np.zeros(dim)
    for i in range(dim - 1):
        out[i] = np.sqrt(((x[i + 1] - x[i] ** 2) ** 2) * 100 + ((x[i] - 1) ** 2))
    return out


@mark.least_squares
def rosenbrock_function_value(params: PyTree) -> FunctionValue:
    """FunctionValue version of the rosenbrock function."""
    return FunctionValue(rosenbrock_ls(params))


@mark.least_squares
def sos_ls(params: PyTree) -> NDArray[np.float64]:
    """Least-squares version of the sum of squares or sphere function."""
    return _get_x(params)


@mark.least_squares
def sos_ls_with_pd_objects(params: PyTree) -> pd.Series[float]:
    """Least-squares version of the sphere function returning pandas objects."""
    return pd.Series(sos_ls(params))


@mark.scalar
def sos_scalar(params: PyTree) -> float:
    """Sum of squares or sphere function."""
    return (_get_x(params) ** 2).sum()


@mark.scalar
def sos_gradient(params: PyTree) -> PyTree:
    """Calculate the gradient of the sum of squares function."""
    flat = 2 * _get_x(params)
    return _unflatten_gradient(flat, params)


@mark.likelihood
def sos_likelihood(params: PyTree) -> NDArray[np.float64]:
    return _get_x(params) ** 2


@mark.likelihood
def sos_likelihood_jacobian(params: PyTree) -> PyTree:
    """Calculate the likelihood Jacobian of the sum of squares function."""
    x = _get_x(params)
    out_mat = np.diag(2 * x)
    out_tree = matrix_to_block_tree(out_mat, x, params)
    return out_tree


@mark.least_squares
def sos_ls_jacobian(params: PyTree) -> PyTree:
    """Calculate the least-squares Jacobian of the sum of squares function."""
    x = _get_x(params)
    out_mat = np.eye(len(x))
    out_tree = matrix_to_block_tree(out_mat, x, params)
    return out_tree


@mark.scalar
def sos_fun_and_gradient(params: PyTree) -> tuple[float, PyTree]:
    """Calculate sum of squares criterion value and gradient."""
    return sos_scalar(params), sos_gradient(params)


@mark.likelihood
def sos_likelihood_fun_and_jac(
    params: PyTree,
) -> tuple[NDArray[np.float64], PyTree]:
    """Calculate sum of squares criterion value and Jacobian."""
    return sos_likelihood(params), sos_likelihood_jacobian(params)


@mark.least_squares
def sos_ls_fun_and_jac(
    params: PyTree,
) -> tuple[NDArray[np.float64], PyTree]:
    """Calculate sum of squares criterion value and Jacobian."""
    return sos_ls(params), sos_ls_jacobian(params)


sos_derivatives = [sos_gradient, sos_likelihood_jacobian, sos_ls_jacobian]


def _get_x(params: PyTree) -> NDArray[np.float64]:
    if isinstance(params, np.ndarray) and params.ndim == 1:
        x = params.astype(float)
    else:
        registry = get_registry(extended=True)
        x = np.array(tree_just_flatten(params, registry=registry), dtype=np.float64)
    return x


def _unflatten_gradient(flat: NDArray[np.float64], params: PyTree) -> PyTree:
    out = tree_unflatten(params, flat.tolist(), registry=REGISTRY)
    return out



from pathlib import Path

import pandas as pd
import plotly.express as px
from packaging import version

DOCS_DIR = Path(__file__).parent.parent / "docs"
OPTIMAGIC_ROOT = Path(__file__).parent

PLOTLY_TEMPLATE = "simple_white"
PLOTLY_PALETTE = px.colors.qualitative.Set2

DEFAULT_N_CORES = 1

CRITERION_PENALTY_SLOPE = 0.1
CRITERION_PENALTY_CONSTANT = 100

# ======================================================================================
# Check Available Packages
# ======================================================================================

try:
    from petsc4py import PETSc  # noqa: F401
except ImportError:
    IS_PETSC4PY_INSTALLED = False
else:
    IS_PETSC4PY_INSTALLED = True

try:
    import nlopt  # noqa: F401
except ImportError:
    IS_NLOPT_INSTALLED = False
else:
    IS_NLOPT_INSTALLED = True

try:
    import pybobyqa  # noqa: F401
except ImportError:
    IS_PYBOBYQA_INSTALLED = False
else:
    IS_PYBOBYQA_INSTALLED = True

try:
    import dfols  # noqa: F401
except ImportError:
    IS_DFOLS_INSTALLED = False
else:
    IS_DFOLS_INSTALLED = True

try:
    import pygmo  # noqa: F401
except ImportError:
    IS_PYGMO_INSTALLED = False
else:
    IS_PYGMO_INSTALLED = True

try:
    import cyipopt  # noqa: F401
except ImportError:
    IS_CYIPOPT_INSTALLED = False
else:
    IS_CYIPOPT_INSTALLED = True

try:
    import fides  # noqa: F401
except ImportError:
    IS_FIDES_INSTALLED = False
else:
    IS_FIDES_INSTALLED = True

try:
    import jax  # noqa: F401
except ImportError:
    IS_JAX_INSTALLED = False
else:
    IS_JAX_INSTALLED = True


try:
    import tranquilo  # noqa: F401
except ImportError:
    IS_TRANQUILO_INSTALLED = False
else:
    IS_TRANQUILO_INSTALLED = True


try:
    import numba  # noqa: F401
except ImportError:
    IS_NUMBA_INSTALLED = False
else:
    IS_NUMBA_INSTALLED = True


# ======================================================================================
# Check if pandas version is newer or equal to version 2.1.0
# ======================================================================================

IS_PANDAS_VERSION_NEWER_OR_EQUAL_TO_2_1_0 = version.parse(
    pd.__version__
) >= version.parse("2.1.0")



"""Implement scipy algorithms.

The following ``scipy`` algorithms are not supported because they
require the specification of the Hessian:

- dogleg
- trust-ncg
- trust-exact
- trust-krylov

The following arguments are not supported as part of ``algo_options``:

- ``disp``
    If set to True would print a convergence message.
    In optimagic it's always set to its default False.
    Refer to optimagic's result dictionary's "success" entry for the convergence
    message.
- ``return_all``
    If set to True, a list of the best solution at each iteration is returned.
    In optimagic it's always set to its default False.
- ``tol``
    This argument of minimize (not an options key) is passed as different types of
    tolerance (gradient, parameter or criterion, as well as relative or absolute)
    depending on the selected algorithm. We require the user to explicitely input
    the tolerance criteria or use our defaults instead.
- ``args``
    This argument of minimize (not an options key) is partialed into the function
    for the user. Specify ``criterion_kwargs`` in ``maximize`` or ``minimize`` to
    achieve the same behavior.
- ``callback``
    This argument would be called after each iteration and the algorithm would
    terminate if it returned True.

"""

import functools
from dataclasses import dataclass
from typing import Any, Callable, List, Literal, Tuple

import numpy as np
import scipy
import scipy.optimize
from numpy.typing import NDArray
from scipy.optimize import Bounds as ScipyBounds
from scipy.optimize import NonlinearConstraint
from scipy.optimize import OptimizeResult as ScipyOptimizeResult

from optimagic import mark
from optimagic.batch_evaluators import process_batch_evaluator
from optimagic.optimization.algo_options import (
    CONVERGENCE_FTOL_ABS,
    CONVERGENCE_FTOL_REL,
    CONVERGENCE_GTOL_ABS,
    CONVERGENCE_GTOL_REL,
    CONVERGENCE_SECOND_BEST_FTOL_ABS,
    CONVERGENCE_SECOND_BEST_XTOL_ABS,
    CONVERGENCE_XTOL_ABS,
    CONVERGENCE_XTOL_REL,
    LIMITED_MEMORY_STORAGE_LENGTH,
    MAX_LINE_SEARCH_STEPS,
    STOPPING_MAXFUN,
    STOPPING_MAXFUN_GLOBAL,
    STOPPING_MAXITER,
)
from optimagic.optimization.algorithm import Algorithm, InternalOptimizeResult
from optimagic.optimization.internal_optimization_problem import (
    InternalBounds,
    InternalOptimizationProblem,
)
from optimagic.parameters.nonlinear_constraints import (
    equality_as_inequality_constraints,
    vector_as_list_of_scalar_constraints,
)
from optimagic.typing import (
    AggregationLevel,
    BatchEvaluator,
    NegativeFloat,
    NonNegativeFloat,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
)
from optimagic.utilities import calculate_trustregion_initial_radius


@mark.minimizer(
    name="scipy_lbfgsb",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyLBFGSB(Algorithm):
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_gtol_abs: NonNegativeFloat = CONVERGENCE_GTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    stopping_maxiter: PositiveInt = STOPPING_MAXITER
    limited_memory_storage_length: PositiveInt = LIMITED_MEMORY_STORAGE_LENGTH
    max_line_search_steps: PositiveInt = MAX_LINE_SEARCH_STEPS

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        options = {
            "maxcor": self.limited_memory_storage_length,
            "ftol": self.convergence_ftol_rel,
            "gtol": self.convergence_gtol_abs,
            "maxfun": self.stopping_maxfun,
            "maxiter": self.stopping_maxiter,
            "maxls": self.max_line_search_steps,
        }
        raw_res = scipy.optimize.minimize(
            fun=problem.fun_and_jac,
            x0=x0,
            method="L-BFGS-B",
            jac=True,
            bounds=_get_scipy_bounds(problem.bounds),
            options=options,
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_slsqp",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=True,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipySLSQP(Algorithm):
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_SECOND_BEST_FTOL_ABS
    stopping_maxiter: PositiveInt = STOPPING_MAXITER

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        options = {
            "maxiter": self.stopping_maxiter,
            "ftol": self.convergence_ftol_abs,
        }
        raw_res = scipy.optimize.minimize(
            fun=problem.fun_and_jac,
            x0=x0,
            method="SLSQP",
            jac=True,
            bounds=_get_scipy_bounds(problem.bounds),
            constraints=problem.nonlinear_constraints,
            options=options,
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_neldermead",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyNelderMead(Algorithm):
    stopping_maxiter: PositiveInt = STOPPING_MAXITER
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_SECOND_BEST_FTOL_ABS
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_SECOND_BEST_XTOL_ABS
    adaptive: bool = False

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        options = {
            "maxiter": self.stopping_maxiter,
            "maxfev": self.stopping_maxfun,
            "xatol": self.convergence_xtol_abs,
            "fatol": self.convergence_ftol_abs,
            # TODO: Benchmark if adaptive = True works better
            "adaptive": self.adaptive,
        }
        raw_res = scipy.optimize.minimize(
            fun=problem.fun,
            x0=x0,
            bounds=_get_scipy_bounds(problem.bounds),
            method="Nelder-Mead",
            options=options,
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_powell",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyPowell(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    stopping_maxiter: PositiveInt = STOPPING_MAXITER

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        options = {
            "xtol": self.convergence_xtol_rel,
            "ftol": self.convergence_ftol_rel,
            "maxfev": self.stopping_maxfun,
            "maxiter": self.stopping_maxiter,
        }
        raw_res = scipy.optimize.minimize(
            fun=problem.fun,
            x0=x0,
            method="Powell",
            bounds=_get_scipy_bounds(problem.bounds),
            options=options,
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_bfgs",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=False,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyBFGS(Algorithm):
    convergence_gtol_abs: NonNegativeFloat = CONVERGENCE_GTOL_ABS
    stopping_maxiter: PositiveInt = STOPPING_MAXITER
    norm: NonNegativeFloat = np.inf

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        options = {
            "gtol": self.convergence_gtol_abs,
            "maxiter": self.stopping_maxiter,
            "norm": self.norm,
        }
        raw_res = scipy.optimize.minimize(
            fun=problem.fun_and_jac, x0=x0, method="BFGS", jac=True, options=options
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_conjugate_gradient",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=False,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyConjugateGradient(Algorithm):
    convergence_gtol_abs: NonNegativeFloat = CONVERGENCE_GTOL_ABS
    stopping_maxiter: PositiveInt = STOPPING_MAXITER
    norm: NonNegativeFloat = np.inf

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        options = {
            "gtol": self.convergence_gtol_abs,
            "maxiter": self.stopping_maxiter,
            "norm": self.norm,
        }
        raw_res = scipy.optimize.minimize(
            fun=problem.fun_and_jac, x0=x0, method="CG", jac=True, options=options
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_newton_cg",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=False,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyNewtonCG(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    stopping_maxiter: PositiveInt = STOPPING_MAXITER

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        options = {
            "xtol": self.convergence_xtol_rel,
            "maxiter": self.stopping_maxiter,
        }
        raw_res = scipy.optimize.minimize(
            fun=problem.fun_and_jac,
            x0=x0,
            method="Newton-CG",
            jac=True,
            options=options,
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_cobyla",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=False,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=True,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyCOBYLA(Algorithm):
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    stopping_maxiter: PositiveInt = STOPPING_MAXITER
    trustregion_initial_radius: PositiveFloat | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        # TODO: Maybe we should leave the radius at their default
        if self.trustregion_initial_radius is None:
            radius = calculate_trustregion_initial_radius(x0)
        else:
            radius = self.trustregion_initial_radius

        options = {
            "maxiter": self.stopping_maxiter,
            "rhobeg": radius,
        }

        # cannot handle equality constraints
        nonlinear_constraints = equality_as_inequality_constraints(
            problem.nonlinear_constraints
        )

        raw_res = scipy.optimize.minimize(
            fun=problem.fun,
            x0=x0,
            method="COBYLA",
            constraints=nonlinear_constraints,
            options=options,
            tol=self.convergence_xtol_rel,
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_ls_trf",
    solver_type=AggregationLevel.LEAST_SQUARES,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyLSTRF(Algorithm):
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_gtol_rel: NonNegativeFloat = CONVERGENCE_GTOL_REL
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    relative_step_size_diff_approx: NonNegativeFloat | None = None
    tr_solver: Literal["exact", "lsmr"] | None = None
    tr_solver_options: dict[str, Any] | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if self.tr_solver_options is None:
            tr_solver_options = {}
        else:
            tr_solver_options = self.tr_solver_options

        raw_res = scipy.optimize.least_squares(
            fun=problem.fun,
            x0=x0,
            # This optimizer does not work with fun_and_jac
            jac=problem.jac,
            bounds=(problem.bounds.lower, problem.bounds.upper),
            method="trf",
            max_nfev=self.stopping_maxfun,
            ftol=self.convergence_ftol_rel,
            gtol=self.convergence_gtol_rel,
            diff_step=self.relative_step_size_diff_approx,
            tr_solver=self.tr_solver,
            tr_options=tr_solver_options,
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_ls_dogbox",
    solver_type=AggregationLevel.LEAST_SQUARES,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyLSDogbox(Algorithm):
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_gtol_rel: NonNegativeFloat = CONVERGENCE_GTOL_REL
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    relative_step_size_diff_approx: NonNegativeFloat | None = None
    tr_solver: Literal["exact", "lsmr"] | None = None
    tr_solver_options: dict[str, Any] | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if self.tr_solver_options is None:
            tr_solver_options = {}
        else:
            tr_solver_options = self.tr_solver_options

        raw_res = scipy.optimize.least_squares(
            fun=problem.fun,
            x0=x0,
            # This optimizer does not work with fun_and_jac
            jac=problem.jac,
            bounds=(problem.bounds.lower, problem.bounds.upper),
            method="dogbox",
            max_nfev=self.stopping_maxfun,
            ftol=self.convergence_ftol_rel,
            gtol=self.convergence_gtol_rel,
            diff_step=self.relative_step_size_diff_approx,
            tr_solver=self.tr_solver,
            tr_options=tr_solver_options,
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_ls_lm",
    solver_type=AggregationLevel.LEAST_SQUARES,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=False,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyLSLM(Algorithm):
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_gtol_rel: NonNegativeFloat = CONVERGENCE_GTOL_REL
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    relative_step_size_diff_approx: NonNegativeFloat | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        raw_res = scipy.optimize.least_squares(
            fun=problem.fun,
            x0=x0,
            # This optimizer does not work with fun_and_jac
            jac=problem.jac,
            method="lm",
            max_nfev=self.stopping_maxfun,
            ftol=self.convergence_ftol_rel,
            gtol=self.convergence_gtol_rel,
            diff_step=self.relative_step_size_diff_approx,
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_truncated_newton",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyTruncatedNewton(Algorithm):
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_gtol_abs: NonNegativeFloat = CONVERGENCE_GTOL_ABS
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    max_hess_evaluations_per_iteration: int = -1
    max_step_for_line_search: NonNegativeFloat = 0
    line_search_severity: float = -1
    finite_difference_precision: NonNegativeFloat = 0
    criterion_rescale_factor: float = -1
    # TODO: Check type hint for `func_min_estimate`
    func_min_estimate: float = 0

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        options = {
            "ftol": self.convergence_ftol_abs,
            "xtol": self.convergence_xtol_abs,
            "gtol": self.convergence_gtol_abs,
            "maxfun": self.stopping_maxfun,
            "maxCGit": self.max_hess_evaluations_per_iteration,
            "stepmx": self.max_step_for_line_search,
            "minfev": self.func_min_estimate,
            "eta": self.line_search_severity,
            "accuracy": self.finite_difference_precision,
            "rescale": self.criterion_rescale_factor,
        }

        raw_res = scipy.optimize.minimize(
            fun=problem.fun_and_jac,
            x0=x0,
            method="TNC",
            jac=True,
            bounds=_get_scipy_bounds(problem.bounds),
            options=options,
        )
        res = process_scipy_result(raw_res)
        return res


@mark.minimizer(
    name="scipy_trust_constr",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=True,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyTrustConstr(Algorithm):
    # TODO: Check if can be set to CONVERGENCE_GTOL_ABS
    convergence_gtol_abs: NonNegativeFloat = 1e-08
    convergence_xtol_rel: NonNegativeFloat = CONVERGENCE_XTOL_REL
    stopping_maxiter: PositiveInt = STOPPING_MAXITER
    trustregion_initial_radius: PositiveFloat | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if self.trustregion_initial_radius is None:
            trustregion_initial_radius = calculate_trustregion_initial_radius(x0)
        else:
            trustregion_initial_radius = self.trustregion_initial_radius

        options = {
            "gtol": self.convergence_gtol_abs,
            "maxiter": self.stopping_maxiter,
            "xtol": self.convergence_xtol_rel,
            "initial_tr_radius": trustregion_initial_radius,
        }

        # cannot handle equality constraints
        nonlinear_constraints = equality_as_inequality_constraints(
            problem.nonlinear_constraints
        )

        raw_res = scipy.optimize.minimize(
            fun=problem.fun_and_jac,
            jac=True,
            x0=x0,
            method="trust-constr",
            bounds=_get_scipy_bounds(problem.bounds),
            constraints=_get_scipy_constraints(nonlinear_constraints),
            options=options,
        )
        res = process_scipy_result(raw_res)
        return res


def process_scipy_result(scipy_res: ScipyOptimizeResult) -> InternalOptimizeResult:
    res = InternalOptimizeResult(
        x=scipy_res.x,
        fun=scipy_res.fun,
        success=bool(scipy_res.success),
        message=str(scipy_res.message),
        n_fun_evals=scipy_res.get("nfev"),
        n_jac_evals=scipy_res.get("njev"),
        n_hess_evals=scipy_res.get("nhev"),
        n_iterations=scipy_res.get("nit"),
        # TODO: Pass on more things once we can convert them to external
        status=None,
        jac=None,
        hess=None,
        hess_inv=None,
        max_constraint_violation=None,
        info=None,
        history=None,
    )
    return res


def _get_scipy_constraints(constraints):
    """Transform internal nonlinear constraints to scipy readable format.

    This format is currently only used by scipy_trust_constr.

    """
    scipy_constraints = [_internal_to_scipy_constraint(c) for c in constraints]
    return scipy_constraints


def _internal_to_scipy_constraint(c):
    new_constr = NonlinearConstraint(
        fun=c["fun"],
        lb=np.zeros(c["n_constr"]),
        ub=np.tile(np.inf, c["n_constr"]),
        jac=c["jac"],
    )
    return new_constr


@mark.minimizer(
    name="scipy_basinhopping",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=True,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyBasinhopping(Algorithm):
    local_algorithm: (
        Literal[
            "Nelder-Mead",
            "Powell",
            "CG",
            "BFGS",
            "Newton-CG",
            "L-BFGS-B",
            "TNC",
            "COBYLA",
            "SLSQP",
            "trust-constr",
            "dogleg",
            "trust-ncg",
            "trust-exact",
            "trust-krylov",
        ]
        | Callable
    ) = "L-BFGS-B"
    n_local_optimizations: PositiveInt = 100
    temperature: NonNegativeFloat = 1.0
    stepsize: NonNegativeFloat = 0.5
    local_algo_options: dict[str, Any] | None = None
    take_step: Callable | None = None
    accept_test: Callable | None = None
    interval: PositiveInt = 50
    convergence_n_unchanged_iterations: PositiveInt | None = None
    seed: int | np.random.Generator | np.random.RandomState | None = None
    target_accept_rate: NonNegativeFloat = 0.5
    stepwise_factor: NonNegativeFloat = 0.9

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        n_local_optimizations = max(1, self.n_local_optimizations - 1)
        if self.local_algo_options is None:
            local_algo_options = {}
        else:
            local_algo_options = self.local_algo_options
        minimizer_kwargs = {
            "method": self.local_algorithm,
            "bounds": _get_scipy_bounds(problem.bounds),
            "jac": problem.jac,
        }
        minimizer_kwargs = {**minimizer_kwargs, **local_algo_options}

        res = scipy.optimize.basinhopping(
            func=problem.fun,
            x0=x0,
            minimizer_kwargs=minimizer_kwargs,
            niter=n_local_optimizations,
            T=self.temperature,
            stepsize=self.stepsize,
            take_step=self.take_step,
            accept_test=self.accept_test,
            interval=self.interval,
            niter_success=self.convergence_n_unchanged_iterations,
            seed=self.seed,
            target_accept_rate=self.target_accept_rate,
            stepwise_factor=self.stepwise_factor,
        )

        return process_scipy_result(res)


@mark.minimizer(
    name="scipy_brute",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=True,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=True,
)
@dataclass(frozen=True)
class ScipyBrute(Algorithm):
    n_grid_points: PositiveInt = 20
    polishing_function: Callable | None = None
    n_cores: PositiveInt = 1
    batch_evaluator: Literal["joblib", "pathos"] | BatchEvaluator = "joblib"

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        workers = _get_workers(self.n_cores, self.batch_evaluator)
        if problem.bounds.lower is None or problem.bounds.upper is None:
            raise ValueError(
                """Global algorithms like scipy_brute need finite bounds
                 for all parameters"""
            )
        raw_res = scipy.optimize.brute(
            func=problem.fun,
            ranges=tuple(
                map(
                    tuple, np.column_stack((problem.bounds.lower, problem.bounds.upper))
                )
            ),
            Ns=self.n_grid_points,
            full_output=True,
            finish=self.polishing_function,
            workers=workers,
        )
        res = InternalOptimizeResult(
            x=raw_res[0],
            fun=raw_res[1],
            n_fun_evals=raw_res[2].size,
            n_iterations=raw_res[2].size,
            success=True,
            message="brute force optimization terminated successfully",
        )

        return res


@mark.minimizer(
    name="scipy_differential_evolution",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=True,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=True,
    disable_history=True,
)
@dataclass(frozen=True)
class ScipyDifferentialEvolution(Algorithm):
    strategy: (
        Literal[
            "best1bin",
            "best1exp",
            "rand1exp",
            "randtobest1exp",
            "currenttobest1exp",
            "best2exp",
            "rand2exp",
            "randtobest1bin",
            "currenttobest1bin",
            "best2bin",
            "rand2bin",
            "rand1bin",
        ]
        | Callable
    ) = "best1bin"
    stopping_maxiter: PositiveInt = STOPPING_MAXFUN_GLOBAL
    population_size_multiplier: NonNegativeInt = 15
    convergence_ftol_rel: NonNegativeFloat = 0.01
    # TODO: Refine type to add ranges [0,2] if float.
    mutation_constant: NonNegativeFloat | Tuple[NonNegativeFloat, NonNegativeFloat] = (
        0.5,
        1,
    )
    # TODO: Refine type to add ranges [0,1].
    recombination_constant: NonNegativeFloat = 0.7
    seed: int | np.random.Generator | np.random.RandomState | None = None
    polish: bool = True
    sampling_method: (
        Literal["latinhypercube", "random", "sobol", "halton"] | NDArray[np.float64]
    ) = "latinhypercube"
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_SECOND_BEST_FTOL_ABS
    n_cores: PositiveInt = 1
    batch_evaluator: Literal["joblib", "pathos"] | BatchEvaluator = "joblib"

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        workers = _get_workers(self.n_cores, self.batch_evaluator)
        res = scipy.optimize.differential_evolution(
            func=problem.fun,
            bounds=_get_scipy_bounds(problem.bounds),
            strategy=self.strategy,
            maxiter=self.stopping_maxiter,
            popsize=self.population_size_multiplier,
            tol=self.convergence_ftol_rel,
            mutation=self.mutation_constant,
            recombination=self.recombination_constant,
            seed=self.seed,
            polish=self.polish,
            init=self.sampling_method,
            atol=self.convergence_ftol_abs,
            updating="deferred",
            workers=workers,
            constraints=_get_scipy_constraints(problem.nonlinear_constraints),
        )

        return process_scipy_result(res)


@mark.minimizer(
    name="scipy_shgo",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=True,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=True,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipySHGO(Algorithm):
    local_algorithm: (
        Literal[
            "Nelder-Mead",
            "Powell",
            "CG",
            "BFGS",
            "Newton-CG",
            "L-BFGS-B",
            "TNC",
            "COBYLA",
            "SLSQP",
            "trust-constr",
            "dogleg",
            "trust-ncg",
            "trust-exact",
            "trust-krylov",
        ]
        | Callable
    ) = "L-BFGS-B"
    local_algo_options: dict[str, Any] | None = None
    n_sampling_points: PositiveInt = 128
    n_simplex_iterations: PositiveInt = 1
    sampling_method: Literal["simplicial", "halton", "sobol"] | Callable = "simplicial"
    max_sampling_evaluations: PositiveInt | None = None
    convergence_minimum_criterion_value: float | None = None
    convergence_minimum_criterion_tolerance: NonNegativeFloat = 1e-4
    stopping_maxiter: PositiveInt | None = None
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN_GLOBAL
    stopping_max_processing_time: PositiveFloat | None = None
    minimum_homology_group_rank_differential: PositiveInt | None = None
    symmetry: List | bool = False
    minimize_every_iteration: bool = True
    max_local_minimizations_per_iteration: PositiveInt | bool = False
    infinity_constraints: bool = True

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if self.local_algorithm == "COBYLA":
            nonlinear_constraints = equality_as_inequality_constraints(
                problem.nonlinear_constraints
            )

        nonlinear_constraints = vector_as_list_of_scalar_constraints(
            problem.nonlinear_constraints
        )

        local_algo_options = (
            {} if self.local_algo_options is None else self.local_algo_options
        )
        default_minimizer_kwargs = {
            "method": self.local_algorithm,
            "bounds": _get_scipy_bounds(problem.bounds),
            "jac": problem.jac,
        }

        minimizer_kwargs = {**default_minimizer_kwargs, **local_algo_options}
        options = {
            "maxfev": self.max_sampling_evaluations,
            "f_min": self.convergence_minimum_criterion_value,
            "f_tol": self.convergence_minimum_criterion_tolerance,
            "maxiter": self.stopping_maxiter,
            "maxev": self.stopping_maxfun,
            "maxtime": self.stopping_max_processing_time,
            "minhgrd": self.minimum_homology_group_rank_differential,
            "symmetry": self.symmetry,
            "jac": problem.jac,
            "minimize_every_iter": self.minimize_every_iteration,
            "local_iter": self.max_local_minimizations_per_iteration,
            "infty_constraints": self.infinity_constraints,
        }

        if any(options.values()) is False:
            options_used = None
        else:
            options_used = options

        res = scipy.optimize.shgo(
            func=problem.fun,
            bounds=_get_scipy_bounds(problem.bounds),
            constraints=nonlinear_constraints,
            minimizer_kwargs=minimizer_kwargs,
            n=self.n_sampling_points,
            iters=self.n_simplex_iterations,
            sampling_method=self.sampling_method,
            options=options_used,
        )

        return process_scipy_result(res)


@mark.minimizer(
    name="scipy_dual_annealing",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=True,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyDualAnnealing(Algorithm):
    stopping_maxiter: PositiveInt = STOPPING_MAXFUN_GLOBAL
    local_algorithm: (
        Literal[
            "Nelder-Mead",
            "Powell",
            "CG",
            "BFGS",
            "Newton-CG",
            "L-BFGS-B",
            "TNC",
            "COBYLA",
            "SLSQP",
            "trust-constr",
            "dogleg",
            "trust-ncg",
            "trust-exact",
            "trust-krylov",
        ]
        | Callable
    ) = "L-BFGS-B"
    local_algo_options: dict[str, Any] | None = None
    # TODO: Refine type to add ranges (0.01, 5e4]
    initial_temperature: PositiveFloat = 5230.0
    # TODO: Refine type to add ranges (0,1)
    restart_temperature_ratio: PositiveFloat = 2e-05
    # TODO: Refine type to add ranges (1, 3]
    visit: PositiveFloat = 2.62
    # TODO: Refine type to add ranges (-1e4, -5]
    accept: NegativeFloat = -5.0
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    seed: int | np.random.Generator | np.random.RandomState | None = None
    no_local_search: bool = False

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        local_algo_options = (
            {} if self.local_algo_options is None else self.local_algo_options
        )
        default_minimizer_kwargs = {
            "method": self.local_algorithm,
            "bounds": _get_scipy_bounds(problem.bounds),
            "jac": problem.jac,
        }

        minimizer_kwargs = {**default_minimizer_kwargs, **local_algo_options}

        res = scipy.optimize.dual_annealing(
            func=problem.fun,
            bounds=_get_scipy_bounds(problem.bounds),
            maxiter=self.stopping_maxiter,
            minimizer_kwargs=minimizer_kwargs,
            initial_temp=self.initial_temperature,
            restart_temp_ratio=self.restart_temperature_ratio,
            visit=self.visit,
            accept=self.accept,
            maxfun=self.stopping_maxfun,
            seed=self.seed,
            no_local_search=self.no_local_search,
            x0=x0,
        )

        return process_scipy_result(res)


@mark.minimizer(
    name="scipy_direct",
    solver_type=AggregationLevel.SCALAR,
    is_available=True,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class ScipyDirect(Algorithm):
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    stopping_maxiter: PositiveInt = STOPPING_MAXFUN_GLOBAL
    locally_biased: bool = True
    convergence_minimum_criterion_value: float = -np.inf
    # TODO: must be between 0 and 1
    convergence_minimum_criterion_tolerance: NonNegativeFloat = 1e-4
    # TODO: must be between 0 and 1
    volume_hyperrectangle_tolerance: NonNegativeFloat = 1e-16
    # TODO: must be between 0 and 1
    length_hyperrectangle_tolerance: NonNegativeFloat = 1e-6

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = scipy.optimize.direct(
            func=problem.fun,
            bounds=_get_scipy_bounds(problem.bounds),
            eps=self.convergence_ftol_rel,
            maxfun=self.stopping_maxfun,
            maxiter=self.stopping_maxiter,
            locally_biased=self.locally_biased,
            f_min=self.convergence_minimum_criterion_value,
            f_min_rtol=self.convergence_minimum_criterion_tolerance,
            vol_tol=self.volume_hyperrectangle_tolerance,
            len_tol=self.length_hyperrectangle_tolerance,
        )

        return process_scipy_result(res)


def _get_workers(n_cores, batch_evaluator):
    batch_evaluator = process_batch_evaluator(batch_evaluator)
    out = functools.partial(
        batch_evaluator,
        n_cores=n_cores,
        error_handling="raise",
    )
    return out


def _get_scipy_bounds(bounds: InternalBounds) -> ScipyBounds:
    return ScipyBounds(lb=bounds.lower, ub=bounds.upper)


def process_scipy_result_old(scipy_results_obj):
    # using get with defaults to access dict elements is just a safety measure
    raw_res = {**scipy_results_obj}
    processed = {
        "solution_x": raw_res.get("x"),
        "solution_criterion": raw_res.get("fun"),
        "solution_derivative": raw_res.get("jac"),
        "solution_hessian": raw_res.get("hess"),
        "n_fun_evals": raw_res.get("nfev"),
        "n_jac_evals": raw_res.get("njac") or raw_res.get("njev"),
        "n_iterations": raw_res.get("nit"),
        "success": raw_res.get("success"),
        "reached_convergence_criterion": None,
        "message": raw_res.get("message"),
    }
    return processed



from abc import ABC, abstractmethod
from dataclasses import KW_ONLY, dataclass
from typing import Any, Callable

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike

from optimagic.exceptions import InvalidConstraintError
from optimagic.optimization.algo_options import CONSTRAINTS_ABSOLUTE_TOLERANCE
from optimagic.typing import PyTree


class Constraint(ABC):
    """Base class for all constraints used for subtyping."""

    @abstractmethod
    def _to_dict(self) -> dict[str, Any]:
        pass


def identity_selector(x: PyTree) -> PyTree:
    return x


@dataclass(frozen=True)
class FixedConstraint(Constraint):
    """Constraint that fixes the selected parameters at their starting values.

    Attributes:
        selector: A function that takes as input the parameters and returns the subset
            of parameters to be constrained. By default, all parameters are constrained.

    Raises:
        InvalidConstraintError: If the selector is not callable.

    """

    selector: Callable[[PyTree], PyTree] = identity_selector

    def _to_dict(self) -> dict[str, Any]:
        return {"type": "fixed", "selector": self.selector}

    def __post_init__(self) -> None:
        if not callable(self.selector):
            raise InvalidConstraintError("'selector' must be callable.")


@dataclass(frozen=True)
class IncreasingConstraint(Constraint):
    """Constraint that ensures the selected parameters are increasing.

    Attributes:
        selector: A function that takes as input the parameters and returns the subset
            of parameters to be constrained. By default, all parameters are constrained.

    Raises:
        InvalidConstraintError: If the selector is not callable.

    """

    selector: Callable[[PyTree], PyTree] = identity_selector

    def _to_dict(self) -> dict[str, Any]:
        return {"type": "increasing", "selector": self.selector}

    def __post_init__(self) -> None:
        if not callable(self.selector):
            raise InvalidConstraintError("'selector' must be callable.")


@dataclass(frozen=True)
class DecreasingConstraint(Constraint):
    """Constraint that ensures that the selected parameters are decreasing.

    Attributes:
        selector: A function that takes as input the parameters and returns the subset
            of parameters to be constrained. By default, all parameters are constrained.

    Raises:
        InvalidConstraintError: If the selector is not callable.

    """

    selector: Callable[[PyTree], PyTree] = identity_selector

    def _to_dict(self) -> dict[str, Any]:
        return {"type": "decreasing", "selector": self.selector}

    def __post_init__(self) -> None:
        if not callable(self.selector):
            raise InvalidConstraintError("'selector' must be callable.")


@dataclass(frozen=True)
class EqualityConstraint(Constraint):
    """Constraint that ensures that the selected parameters are equal.

    Attributes:
        selector: A function that takes as input the parameters and returns the subset
            of parameters to be constrained. By default, all parameters are constrained.

    Raises:
        InvalidConstraintError: If the selector is not callable.

    """

    selector: Callable[[PyTree], PyTree] = identity_selector

    def _to_dict(self) -> dict[str, Any]:
        return {"type": "equality", "selector": self.selector}

    def __post_init__(self) -> None:
        if not callable(self.selector):
            raise InvalidConstraintError("'selector' must be callable.")


@dataclass(frozen=True)
class ProbabilityConstraint(Constraint):
    """Constraint that ensures that the selected parameters are probabilities.

    This constraint ensures that each of the selected parameters is positive and that
    the sum of the selected parameters is 1.

    Attributes:
        selector: A function that takes as input the parameters and returns the subset
            of parameters to be constrained. By default, all parameters are constrained.

    Raises:
        InvalidConstraintError: If the selector is not callable.

    """

    selector: Callable[[PyTree], PyTree] = identity_selector

    def _to_dict(self) -> dict[str, Any]:
        return {"type": "probability", "selector": self.selector}

    def __post_init__(self) -> None:
        if not callable(self.selector):
            raise InvalidConstraintError("'selector' must be callable.")


@dataclass(frozen=True)
class PairwiseEqualityConstraint(Constraint):
    """Constraint that ensures that groups of selected parameters are equal.

    This constraint ensures that each pair between the selected parameters is equal.

    Attributes:
        selectors: A list of functions that take as input the parameters and return the
            subsets of parameters to be constrained.

    Raises:
        InvalidConstraintError: If the selector is not callable.

    """

    selectors: list[Callable[[PyTree], PyTree]]

    def _to_dict(self) -> dict[str, Any]:
        return {"type": "pairwise_equality", "selectors": self.selectors}

    def __post_init__(self) -> None:
        if len(self.selectors) < 2:
            raise InvalidConstraintError("At least two selectors must be provided.")

        if not all(callable(s) for s in self.selectors):
            raise InvalidConstraintError("All selectors must be callable.")


@dataclass(frozen=True)
class FlatCovConstraint(Constraint):
    """Constraint that ensures the selected parameters are a valid covariance matrix.

    Attributes:
        selector: A function that takes as input the parameters and returns the subset
            of parameters to be constrained. By default, all parameters are constrained.
        regularization: Helps in guiding the optimization towards finding a
            positive definite covariance matrix instead of only a positive semi-definite
            matrix. Larger values correspond to a higher likelihood of positive
            definiteness. Defaults to 0.

    Raises:
        InvalidConstraintError: If the selector is not callable or regularization is
            not a non-negative float or int.

    """

    selector: Callable[[PyTree], PyTree] = identity_selector
    _: KW_ONLY
    regularization: float = 0.0

    def _to_dict(self) -> dict[str, Any]:
        return {
            "type": "covariance",
            "selector": self.selector,
            "regularization": self.regularization,
        }

    def __post_init__(self) -> None:
        if not callable(self.selector):
            raise InvalidConstraintError("'selector' must be callable.")

        if not isinstance(self.regularization, float | int) or self.regularization < 0:
            raise InvalidConstraintError(
                "'regularization' must be a non-negative float or int."
            )


@dataclass(frozen=True)
class FlatSDCorrConstraint(Constraint):
    """Constraint that ensures the selected parameters are a valid correlation matrix.

    This constraint ensures that each of the selected parameters is positive and that
    the sum of the selected parameters is 1.

    Attributes:
        selector: A function that takes as input the parameters and returns the subset
            of parameters to be constrained. By default, all parameters are constrained.
        regularization: Helps in guiding the optimization towards finding a
            positive definite covariance matrix instead of only a positive semi-definite
            matrix. Larger values correspond to a higher likelihood of positive
            definiteness. Defaults to 0.

    Raises:
        InvalidConstraintError: If the selector is not callable or regularization is
            not a non-negative float or int.

    """

    selector: Callable[[PyTree], PyTree] = identity_selector
    _: KW_ONLY
    regularization: float = 0.0

    def _to_dict(self) -> dict[str, Any]:
        return {
            "type": "sdcorr",
            "selector": self.selector,
            "regularization": self.regularization,
        }

    def __post_init__(self) -> None:
        if not callable(self.selector):
            raise InvalidConstraintError("'selector' must be callable.")

        if not isinstance(self.regularization, float | int) or self.regularization < 0:
            raise InvalidConstraintError(
                "'regularization' must be a non-negative float or int."
            )


@dataclass(frozen=True)
class LinearConstraint(Constraint):
    """Constraint that bounds a linear combination of the selected parameters.

    This constraint ensures that a linear combination of the selected parameters with
    the 'weights' is either equal to 'value', or is bounded by 'lower_bound' and
    'upper_bound'.

    Attributes:
        selector: A function that takes as input the parameters and returns the subset
            of parameters to be constrained. By default, all parameters are constrained.
        weights: The weights for the linear combination. If a scalar is provided, it is
            used for all parameters. Otherwise, it must have the same structure as the
            selected parameters.
        lower_bound: The lower bound for the linear combination. Defaults to None.
        upper_bound: The upper bound for the linear combination. Defaults to None.
        value: The value to compare the linear combination to. Defaults to None.

    Raises:
        InvalidConstraintError: If the selector is not callable, or if the weights,
            lower_bound, upper_bound, or value are not valid.

    """

    selector: Callable[[PyTree], ArrayLike | "pd.Series[float]" | float | int] = (
        identity_selector
    )
    _: KW_ONLY
    weights: ArrayLike | "pd.Series[float]" | float | int | None = None
    lower_bound: float | int | None = None
    upper_bound: float | int | None = None
    value: float | int | None = None

    def _to_dict(self) -> dict[str, Any]:
        return {
            "type": "linear",
            "selector": self.selector,
            "weights": self.weights,
            **_select_non_none(
                lower_bound=self.lower_bound,
                upper_bound=self.upper_bound,
                value=self.value,
            ),
        }

    def __post_init__(self) -> None:
        if not callable(self.selector):
            raise InvalidConstraintError("'selector' must be callable.")

        if _all_none(self.lower_bound, self.upper_bound, self.value):
            raise InvalidConstraintError(
                "At least one of 'lower_bound', 'upper_bound', or 'value' must be "
                "non-None."
            )
        if self.value is not None and not _all_none(self.lower_bound, self.upper_bound):
            raise InvalidConstraintError(
                "'value' cannot be used with 'lower_bound' or 'upper_bound'."
            )

        if not isinstance(self.weights, np.ndarray | list | pd.Series | float | int):
            raise InvalidConstraintError(
                "'weights' must be an array-like, a pandas Series, a float, or an int."
            )

        if self.lower_bound is not None and not isinstance(
            self.lower_bound, float | int
        ):
            raise InvalidConstraintError("'lower_bound' must be a float or an int.")

        if self.upper_bound is not None and not isinstance(
            self.upper_bound, float | int
        ):
            raise InvalidConstraintError("'upper_bound' must be a float or an int.")

        if self.value is not None and not isinstance(self.value, float | int):
            raise InvalidConstraintError("'value' must be a float or an int.")


@dataclass(frozen=True)
class NonlinearConstraint(Constraint):
    """Constraint that bounds a nonlinear function of the selected parameters.

    This constraint ensures that a nonlinear function of the selected parameters is
    either equal to 'value', or is bounded by 'lower_bound' and 'upper_bound'.

    Attributes:
        selector: A function that takes as input the parameters and returns the subset
            of parameters to be constrained. By default, all parameters are constrained.
        func: The constraint function which is applied to the selected parameters.
        derivative: The derivative of the constraint function with respect to the
            selected parameters. Defaults to None.
        lower_bound: The lower bound for the nonlinear function. Can be a scalar or of
            the same structure as output of the constraint function. Defaults to None.
        upper_bound: The upper bound for the nonlinear function. Can be a scalar or of
            the same structure as output of the constraint function. Defaults to None.
        value: The value to compare the nonlinear function to. Can be a scalar or of
            the same structure as output of the constraint function. Defaults to None.
        tol: The tolerance for the constraint function. Defaults to
            `optimagic.optimization.algo_options.CONSTRAINTS_ABSOLUTE_TOLERANCE`.

    Raises:
        InvalidConstraintError: If the selector is not callable, or if the func,
            derivative, lower_bound, upper_bound, or value are not valid.

    """

    selector: Callable[[PyTree], PyTree] = identity_selector
    _: KW_ONLY
    func: Callable[[PyTree], ArrayLike | "pd.Series[float]" | float] | None = None
    derivative: Callable[[PyTree], PyTree] | None = None
    lower_bound: ArrayLike | "pd.Series[float]" | float | None = None
    upper_bound: ArrayLike | "pd.Series[float]" | float | None = None
    value: ArrayLike | "pd.Series[float]" | float | None = None
    tol: float = CONSTRAINTS_ABSOLUTE_TOLERANCE

    def _to_dict(self) -> dict[str, Any]:
        return {
            "type": "nonlinear",
            "selector": self.selector,
            **_select_non_none(
                func=self.func,
                derivative=self.derivative,
                # In the dict representation, we write _bounds instead of _bound.
                lower_bounds=self.lower_bound,
                upper_bounds=self.upper_bound,
                value=self.value,
                tol=self.tol,
            ),
        }

    def __post_init__(self) -> None:
        if not callable(self.selector):
            raise InvalidConstraintError("'selector' must be callable.")

        if _all_none(self.lower_bound, self.upper_bound, self.value):
            raise InvalidConstraintError(
                "At least one of 'lower_bound', 'upper_bound', or 'value' must be "
                "non-None."
            )
        if self.value is not None and not _all_none(self.lower_bound, self.upper_bound):
            raise InvalidConstraintError(
                "'value' cannot be used with 'lower_bound' or 'upper_bound'."
            )

        if self.tol is not None and (
            not isinstance(self.tol, float | int) or self.tol < 0
        ):
            raise InvalidConstraintError("'tol' must be non-negative.")

        if self.func is None or not callable(self.func):
            raise InvalidConstraintError("'func' must be callable.")

        if self.derivative is not None and not callable(self.derivative):
            raise InvalidConstraintError("'derivative' must be callable.")


def _all_none(*args: Any) -> bool:
    return all(v is None for v in args)


def _select_non_none(**kwargs: Any) -> dict[str, Any]:
    return {k: v for k, v in kwargs.items() if v is not None}
